#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
import logging

from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common import cliutils
from manilaclient.osc import utils

LOG = logging.getLogger(__name__)


class CreateShareReplica(command.ShowOne):
    """Create a share replica."""

    _description = _("Create a replica of the given share")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share",
            metavar="<share>",
            help=_("Name or ID of the share to replicate."),
        )
        parser.add_argument(
            '--availability-zone',
            metavar='<availability-zone>',
            default=None,
            help=_(
                'Availability zone in which the replica should be created.'
            ),
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            default=False,
            help=_('Wait for replica creation'),
        )
        parser.add_argument(
            "--scheduler-hint",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Scheduler hint for the share replica as key=value pairs, "
                "Supported key is only_host. Available for microversion "
                ">= 2.67."
            ),
        )
        parser.add_argument(
            '--share-network',
            metavar='<share-network-name-or-id>',
            default=None,
            help=_(
                'Optional network info ID or name. Available for '
                'microversion >= 2.72'
            ),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this replica "
                "(repeat option to set multiple properties). "
                "Available only for microversion >= 2.95"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = osc_utils.find_resource(share_client.shares, parsed_args.share)
        scheduler_hints = {}
        if parsed_args.scheduler_hint:
            if share_client.api_version < api_versions.APIVersion("2.67"):
                raise exceptions.CommandError(
                    _(
                        "arg '--scheduler_hint' is available only starting "
                        "with API microversion '2.67'."
                    )
                )

            hints = utils.extract_key_value_options(parsed_args.scheduler_hint)
            if 'only_host' not in hints.keys() or len(hints) > 1:
                raise exceptions.CommandError(
                    "The only valid key supported with the --scheduler-hint "
                    "argument is 'only_host'."
                )
            scheduler_hints['only_host'] = hints.get('only_host')

        body = {
            'share': share,
            'availability_zone': parsed_args.availability_zone,
        }

        metadata = {}
        if parsed_args.property:
            if share_client.api_version < api_versions.APIVersion("2.95"):
                raise exceptions.CommandError(
                    _(
                        "arg '--property' is available only starting "
                        "with API microversion '2.95'."
                    )
                )
            metadata = parsed_args.property

        if metadata:
            body['metadata'] = metadata

        if scheduler_hints:
            body['scheduler_hints'] = scheduler_hints

        share_network_id = None
        if parsed_args.share_network:
            if share_client.api_version < api_versions.APIVersion("2.72"):
                raise exceptions.CommandError(
                    "'share-network' option is available only starting "
                    "with '2.72' API microversion."
                )
            share_network_id = osc_utils.find_resource(
                share_client.share_networks, parsed_args.share_network
            ).id
            body['share_network'] = share_network_id

        share_replica = share_client.share_replicas.create(**body)
        if parsed_args.wait:
            if not osc_utils.wait_for_status(
                status_f=share_client.share_replicas.get,
                res_id=share_replica.id,
                success_status=['available'],
            ):
                LOG.error(_("ERROR: Share replica is in error state."))

            share_replica = osc_utils.find_resource(
                share_client.share_replicas, share_replica.id
            )

        return self.dict2columns(share_replica._info)


class DeleteShareReplica(command.Command):
    """Delete one or more share replicas."""

    _description = _("Delete one or more share replicas")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "replica",
            metavar="<replica>",
            nargs="+",
            help=_("Name or ID of the replica(s) to delete"),
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_(
                "Attempt to force delete a replica on its backend. "
                "Using this option will purge the replica from Manila "
                "even if it is not cleaned up on the backend. "
            ),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share replica deletion"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for replica in parsed_args.replica:
            try:
                replica_obj = osc_utils.find_resource(
                    share_client.share_replicas, replica
                )

                share_client.share_replicas.delete(
                    replica_obj, force=parsed_args.force
                )

                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                        manager=share_client.share_replicas,
                        res_id=replica_obj.id,
                    ):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete a share replica with "
                        "name or ID '%(replica)s': %(e)s"
                    ),
                    {'replica': replica, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.replica)
            msg = _("%(result)s of %(total)s replicas failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListShareReplica(command.Lister):
    """List share replicas."""

    _description = _("List share replicas")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--share",
            metavar="<share>",
            default=None,
            help=_("Name or ID of the share to list replicas for."),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_(
                'Filter replicas having a given metadata key=value '
                'property. (repeat option to filter by multiple '
                'properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = None
        if parsed_args.share:
            share = osc_utils.find_resource(
                share_client.shares, parsed_args.share
            )

        properties = parsed_args.property or {}
        if properties:
            if share_client.api_version < api_versions.APIVersion("2.95"):
                raise exceptions.CommandError(
                    "Property based filtering is only available "
                    "with manila API version >= 2.95"
                )

        search_opts = {}
        if properties:
            meta_str = ",".join(f"{k}:{v}" for k, v in properties.items())
            search_opts['metadata'] = meta_str

        replicas = share_client.share_replicas.list(
            share=share, search_opts=search_opts
        )

        if properties:
            replicas = [
                r
                for r in replicas
                if all(
                    r._info.get('metadata', {}).get(k) == v
                    for k, v in properties.items()
                )
            ]

        columns = [
            'id',
            'status',
            'replica_state',
            'share_id',
            'host',
            'availability_zone',
            'updated_at',
        ]

        if share_client.api_version >= api_versions.APIVersion("2.95"):
            for replica in replicas:
                replica._info.update(
                    {
                        'properties': format_columns.DictColumn(
                            replica._info.pop('metadata', {})
                        ),
                    },
                )

            columns += ['properties']

        column_headers = utils.format_column_headers(columns)
        data = (
            osc_utils.get_dict_properties(replica._info, columns)
            for replica in replicas
        )

        return (column_headers, data)


class ShowShareReplica(command.ShowOne):
    """Show share replica."""

    _description = _("Show details about a replica")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "replica",
            metavar="<replica>",
            help=_(
                "ID of the share replica. Available only for "
                "microversion >= 2.47. "
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        replica = share_client.share_replicas.get(parsed_args.replica)

        replica_export_locations = (
            share_client.share_replica_export_locations.list(
                share_replica=replica
            )
        )

        replica._info['export_locations'] = []
        for element_location in replica_export_locations:
            element_location._info.pop('links', None)
            replica._info['export_locations'].append(element_location._info)

        if parsed_args.formatter == 'table':
            replica._info['export_locations'] = (
                cliutils.convert_dict_list_to_string(
                    replica._info['export_locations']
                )
            )
        # Special mapping for columns to make the output easier to read:
        # 'metadata' --> 'properties'
        replica._info.update(
            {
                'properties': format_columns.DictColumn(
                    replica._info.pop('metadata', {})
                ),
            },
        )

        replica._info.pop('links', None)

        return self.dict2columns(replica._info)


class SetShareReplica(command.Command):
    """Set share replica"""

    _description = _(
        "Explicitly set share replica status and/or replica-state "
        "and/or property"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "replica",
            metavar="<replica>",
            help=_("ID of the share replica to modify."),
        )
        parser.add_argument(
            "--replica-state",
            metavar="<replica-state>",
            choices=['in_sync', 'out_of_sync', 'active', 'error'],
            help=_(
                "Indicate which replica_state to assign the replica. "
                "Options include in_sync, out_of_sync, active and error."
            ),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'error_deleting',
            ],
            help=_(
                "Indicate which status to assign the replica. Options "
                "include available, error, creating, deleting and "
                "error_deleting."
            ),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this replica "
                "(repeat option to set multiple properties). "
                "Available only for microversion >= 2.95"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        replica = osc_utils.find_resource(
            share_client.share_replicas, parsed_args.replica
        )

        if parsed_args.replica_state:
            try:
                share_client.share_replicas.reset_replica_state(
                    replica, parsed_args.replica_state
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to set replica_state "
                        "'%(replica_state)s': %(exception)s"
                    ),
                    {
                        'replica_state': parsed_args.replica_state,
                        'exception': e,
                    },
                )

        if parsed_args.status:
            try:
                share_client.share_replicas.reset_state(
                    replica, parsed_args.status
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _("Failed to set status '%(status)s': %(exception)s"),
                    {'status': parsed_args.status, 'exception': e},
                )

        if parsed_args.property:
            if share_client.api_version < api_versions.APIVersion("2.95"):
                raise exceptions.CommandError(
                    _(
                        "Setting properties in share replicas is available "
                        "only starting with API microversion '2.95'."
                    )
                )
            try:
                replica.set_metadata(parsed_args.property)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to set share replica properties "
                        "'%(properties)s': %(exception)s"
                    ),
                    {'properties': parsed_args.property, 'exception': e},
                )
                result += 1

        if (
            not parsed_args.replica_state
            and not parsed_args.status
            and not parsed_args.property
        ):
            raise exceptions.CommandError(
                _(
                    "Nothing to set. Please define "
                    "either '--replica_state' or '--status' or '--property'."
                )
            )
        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class UnsetShareReplica(command.Command):
    """Unset a share replica property."""

    _description = _("Unset a share replica property")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "replica",
            metavar="<replica>",
            help=_("ID of the share replica to unset property from"),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            help=_(
                'Remove a property from replica '
                '(repeat option to remove multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        replica = osc_utils.find_resource(
            share_client.share_replicas, parsed_args.replica
        )

        if parsed_args.property:
            if share_client.api_version < api_versions.APIVersion("2.95"):
                raise exceptions.CommandError(
                    _(
                        "arg '--property' is available only starting "
                        "with API microversion '2.95'."
                    )
                )
            result = 0
            for key in parsed_args.property:
                try:
                    replica.delete_metadata([key])
                except Exception as e:
                    result += 1
                    LOG.error(
                        _(
                            "Failed to unset replica property "
                            "'%(key)s': %(exception)s"
                        ),
                        {'key': key, 'exception': e},
                    )
            if result > 0:
                raise exceptions.CommandError(
                    _("One or more of the unset operations failed")
                )
        else:
            raise exceptions.CommandError(
                "Please specify '--property <key>' to unset a property."
            )


class PromoteShareReplica(command.Command):
    """Promote share replica"""

    _description = _("Promote specified replica to 'active' replica_state.")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "replica", metavar="<replica>", help=_("ID of the share replica.")
        )
        parser.add_argument(
            '--quiesce-wait-time',
            metavar='<quiesce-wait-time>',
            default=None,
            help=_(
                'Quiesce wait time in seconds. Available for '
                'microversion >= 2.75'
            ),
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            default=False,
            help=_('Wait for share replica promotion'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        replica = osc_utils.find_resource(
            share_client.share_replicas, parsed_args.replica
        )

        args = [
            replica,
        ]
        if parsed_args.quiesce_wait_time:
            if share_client.api_version < api_versions.APIVersion("2.75"):
                raise exceptions.CommandError(
                    "'quiesce-wait-time' option is available only starting "
                    "with '2.75' API microversion."
                )
            args += [parsed_args.quiesce_wait_time]

        try:
            share_client.share_replicas.promote(*args)
            if parsed_args.wait:
                if not osc_utils.wait_for_status(
                    status_f=share_client.share_replicas.get,
                    res_id=replica.id,
                    success_status=['active'],
                    status_field='replica_state',
                ):
                    LOG.error(_("ERROR: Share replica is in error state."))

        except Exception as e:
            msg = "Failed to promote replica to 'active': %(e)s"
            raise exceptions.CommandError(msg % {'e': e})


class ResyncShareReplica(command.Command):
    """Resync share replica"""

    _description = _(
        "Attempt to update the share replica with its 'active' mirror."
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "replica",
            metavar="<replica>",
            help=_("ID of the share replica to resync."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        replica = osc_utils.find_resource(
            share_client.share_replicas, parsed_args.replica
        )

        try:
            share_client.share_replicas.resync(replica)
        except Exception as e:
            msg = "Failed to resync share replica: %(e)s"
            raise exceptions.CommandError(msg % {'e': e})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base
from manilaclient.common import constants


class ResourceLock(base.Resource):
    """Lock a share resource action from being executed"""

    def __repr__(self):
        return f"<ResourceLock: {self.id}>"

    def delete(self):
        """Delete this lock."""
        return self.manager.delete(self)

    def update(self, **kwargs):
        """Update this lock."""
        return self.manager.update(self, **kwargs)


class ResourceLockManager(base.ManagerWithFind):
    """Manage :class:`ResourceLock` resources."""

    resource_class = ResourceLock

    @api_versions.wraps(constants.RESOURCE_LOCK_VERSION)
    def create(
        self,
        resource_id,
        resource_type,
        resource_action='delete',
        lock_reason=None,
    ):
        """Creates a resource lock.

        :param resource_id: The ID of the resource to lock
        :param resource_type: The type of the resource (e.g., "share",
        "access")
        :param resource_action: The functionality to lock (e.g., "delete",
        "view/delete")
        :param lock_reason: Lock description
        :rtype: :class:`ResourceLock`
        """
        body = {
            'resource_lock': {
                'resource_id': resource_id,
                'resource_type': resource_type,
                'resource_action': resource_action,
                'lock_reason': lock_reason,
            }
        }

        return self._create('/resource-locks', body, 'resource_lock')

    @api_versions.wraps(constants.RESOURCE_LOCK_VERSION)
    def get(self, lock_id):
        """Show details of a resource lock.

        :param lock_id: The ID of the resource lock to display.
        :rtype: :class:`ResourceLock`
        """
        return self._get(f"/resource-locks/{lock_id}", "resource_lock")

    @api_versions.wraps(constants.RESOURCE_LOCK_VERSION)
    def list(self, search_opts=None, sort_key=None, sort_dir=None):
        """Get a list of all resource locks.

        :param search_opts: Filtering options as a dictionary.
        :param sort_key: Key to be sorted (i.e. 'created_at').
        :param sort_dir: Sort direction, should be 'desc' or 'asc'.
        :rtype: list of :class:`ResourceLock`
        """
        search_opts = search_opts or {}

        sort_key = sort_key or 'created_at'
        if sort_key in constants.RESOURCE_LOCK_SORT_KEY_VALUES:
            search_opts['sort_key'] = sort_key
        else:
            raise ValueError(
                'sort_key must be one of the following: {}.'.format(
                    ', '.join(constants.RESOURCE_LOCK_SORT_KEY_VALUES)
                )
            )
        sort_dir = sort_dir or 'desc'
        if sort_dir in constants.SORT_DIR_VALUES:
            search_opts['sort_dir'] = sort_dir
        else:
            raise ValueError(
                'sort_dir must be one of the following: {}.'.format(
                    ', '.join(constants.SORT_DIR_VALUES)
                )
            )

        query_string = self._build_query_string(search_opts)

        path = f"/resource-locks{query_string}"

        return self._list(path, 'resource_locks')

    @api_versions.wraps(constants.RESOURCE_LOCK_VERSION)
    def update(self, lock, **kwargs):
        """Updates a resource lock.

        :param lock: The :class:`ResourceLock` object, or a lock id to update.
        :param kwargs: "resource_action" and "lock_reason" are allowed kwargs
        :rtype: :class:`ResourceLock`
        """
        if not kwargs:
            return

        body = {'resource_lock': {}}

        if 'lock_reason' in kwargs:
            body['resource_lock']['lock_reason'] = kwargs['lock_reason']
        if 'resource_action' in kwargs:
            body['resource_lock']['resource_action'] = kwargs[
                'resource_action'
            ]

        lock_id = base.getid(lock)
        return self._update(f"/resource-locks/{lock_id}", body)

    @api_versions.wraps(constants.RESOURCE_LOCK_VERSION)
    def delete(self, lock):
        """Delete a resource lock.

        :param lock: The :class:`ResourceLock` object, or a lock id to delete.
        """
        return self._delete(f"/resource-locks/{base.getid(lock)}")



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils
from oslo_utils import strutils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import constants
from manilaclient.osc import utils

LOG = logging.getLogger(__name__)

ATTRIBUTES = [
    'id',
    'name',
    'visibility',
    'is_default',
    'required_extra_specs',
    'optional_extra_specs',
    'description',
]


def format_share_type(share_type, formatter='table'):
    # share_type_access:is_public (true/false) --> visibility (public/private)
    is_public = 'share_type_access:is_public'
    visibility = 'public' if share_type._info.get(is_public) else 'private'
    share_type._info.pop(is_public, None)

    # optional_extra_specs --> extra_specs without required_extra_specs
    # required_extra_specs are displayed separately
    optional_extra_specs = share_type.extra_specs
    for key in share_type.required_extra_specs.keys():
        optional_extra_specs.pop(key, None)

    if formatter == 'table':
        share_type._info.update(
            {
                'visibility': visibility,
                'required_extra_specs': utils.format_properties(
                    share_type.required_extra_specs
                ),
                'optional_extra_specs': utils.format_properties(
                    optional_extra_specs
                ),
            }
        )
    else:
        share_type._info.update(
            {
                'visibility': visibility,
                'required_extra_specs': share_type.required_extra_specs,
                'optional_extra_specs': optional_extra_specs,
            }
        )

    return share_type


class CreateShareType(command.ShowOne):
    """Create new share type."""

    _description = _("Create new share type")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar="<name>", default=None, help=_('Share type name')
        )
        parser.add_argument(
            'spec_driver_handles_share_servers',
            metavar="<spec_driver_handles_share_servers>",
            default=None,
            help=_(
                "Required extra specification. "
                "Valid values are 'true' and 'false'"
            ),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_(
                "Share type description. "
                "Available only for microversion >= 2.41."
            ),
        )
        parser.add_argument(
            "--snapshot-support",
            metavar="<snapshot_support>",
            default=None,
            help=_(
                "Boolean extra spec used for filtering of back ends "
                "by their capability to create share snapshots."
            ),
        )
        parser.add_argument(
            "--create-share-from-snapshot-support",
            metavar="<create_share_from_snapshot_support>",
            default=None,
            help=_(
                "Boolean extra spec used for filtering of back ends "
                "by their capability to create shares from snapshots."
            ),
        )
        parser.add_argument(
            "--revert-to-snapshot-support",
            metavar="<revert_to_snapshot_support>",
            default=False,
            help=_(
                "Boolean extra spec used for filtering of back ends "
                "by their capability to revert shares to snapshots. "
                "(Default is False)."
            ),
        )
        parser.add_argument(
            "--mount-snapshot-support",
            metavar="<mount_snapshot_support>",
            default=False,
            help=_(
                "Boolean extra spec used for filtering of back ends "
                "by their capability to mount share snapshots. "
                "(Default is False)."
            ),
        )
        parser.add_argument(
            "--extra-specs",
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_(
                "Extra specs key and value of share type that will be"
                " used for share type creation. OPTIONAL: Default=None."
                " example --extra-specs  thin_provisioning='<is> True', "
                "replication_type=readable."
            ),
        )
        parser.add_argument(
            '--public',
            metavar="<public>",
            default=True,
            help=_('Make type accessible to the public (default true).'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        kwargs = {'name': parsed_args.name}
        try:
            kwargs['spec_driver_handles_share_servers'] = (
                strutils.bool_from_string(
                    parsed_args.spec_driver_handles_share_servers, strict=True
                )
            )
        except ValueError as e:
            msg = (
                "Argument spec_driver_handles_share_servers "
                f"argument is not valid: {str(e)}"
            )
            raise exceptions.CommandError(msg)

        if parsed_args.description:
            if share_client.api_version.matches(
                api_versions.APIVersion("2.41"), api_versions.APIVersion()
            ):
                kwargs['description'] = parsed_args.description
            else:
                raise exceptions.CommandError(
                    "Adding description to share type "
                    "is only available with API microversion >= 2.41"
                )

        if parsed_args.public:
            kwargs['is_public'] = strutils.bool_from_string(
                parsed_args.public, default=True
            )

        extra_specs = {}
        if parsed_args.extra_specs:
            for item in parsed_args.extra_specs:
                (key, value) = item.split('=', 1)
                if key == 'driver_handles_share_servers':
                    msg = (
                        "'driver_handles_share_servers' "
                        "is already set via positional argument."
                    )
                    raise exceptions.CommandError(msg)
                else:
                    extra_specs = utils.extract_extra_specs(
                        extra_specs, [item]
                    )

        for key in constants.BOOL_SPECS:
            value = getattr(parsed_args, key)
            if value:
                extra_specs = utils.extract_extra_specs(
                    extra_specs, [key + '=' + value]
                )

        kwargs['extra_specs'] = extra_specs

        share_type = share_client.share_types.create(**kwargs)
        formatted_type = format_share_type(share_type, parsed_args.formatter)

        return (
            ATTRIBUTES,
            oscutils.get_dict_properties(formatted_type._info, ATTRIBUTES),
        )


class DeleteShareType(command.Command):
    """Delete a share type."""

    _description = _("Delete a share type")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_types',
            metavar="<share_types>",
            nargs="+",
            help=_("Name or ID of the share type(s) to delete"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for share_type in parsed_args.share_types:
            try:
                share_type_obj = apiutils.find_resource(
                    share_client.share_types, share_type
                )

                share_client.share_types.delete(share_type_obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete share type with "
                        "name or ID '%(share_type)s': %(e)s"
                    ),
                    {'share_type': share_type, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.share_types)
            msg = _(
                "%(result)s of %(total)s share types failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class SetShareType(command.Command):
    """Set share type properties."""

    _description = _("Set share type properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_type',
            metavar="<share_type>",
            help=_("Name or ID of the share type to modify"),
        )
        parser.add_argument(
            "--extra-specs",
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_(
                "Extra specs key and value of share type that will be"
                " used for share type creation. OPTIONAL: Default=None."
                " example --extra-specs  thin_provisioning='<is> True', "
                "replication_type=readable."
            ),
        )
        parser.add_argument(
            '--public',
            metavar="<public>",
            default=None,
            help=_(
                'New visibility of the share type. If set to True, '
                'share type will be available to all projects '
                'in the cloud. '
                'Available only for microversion >= 2.50'
            ),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_(
                "New description of share type. "
                "Available only for microversion >= 2.50"
            ),
        )
        parser.add_argument(
            '--name',
            metavar="<name>",
            default=None,
            help=_(
                'New name of share type. '
                'Available only for microversion >= 2.50'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_type = apiutils.find_resource(
            share_client.share_types, parsed_args.share_type
        )

        can_update = share_client.api_version >= api_versions.APIVersion(
            '2.50'
        )

        kwargs = {}
        if parsed_args.name is not None:
            if can_update:
                kwargs['name'] = parsed_args.name
            else:
                raise exceptions.CommandError(
                    "Setting (new) name to share type "
                    "is only available with API microversion >= 2.50"
                )
        if parsed_args.description is not None:
            if can_update:
                kwargs['description'] = parsed_args.description
            else:
                raise exceptions.CommandError(
                    "Setting (new) description to share type "
                    "is only available with API microversion >= 2.50"
                )
        if parsed_args.public is not None:
            if can_update:
                kwargs['is_public'] = strutils.bool_from_string(
                    parsed_args.public, default=True
                )
            else:
                raise exceptions.CommandError(
                    "Setting visibility to share type "
                    "is only available with API microversion >= 2.50"
                )
        if kwargs:
            share_type.update(**kwargs)

        if parsed_args.extra_specs:
            extra_specs = utils.extract_extra_specs(
                extra_specs={}, specs_to_add=parsed_args.extra_specs
            )
            try:
                share_type.set_keys(extra_specs)
            except Exception as e:
                raise exceptions.CommandError(
                    f"Failed to set share type key: {e}"
                )


class UnsetShareType(command.Command):
    """Unset share type extra specs."""

    _description = _("Unset share type extra specs")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_type',
            metavar="<share_type>",
            help=_("Name or ID of the share type to modify"),
        )
        parser.add_argument(
            'extra_specs',
            metavar='<key>',
            nargs='+',
            help=_('Remove extra_specs from this share type'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_type = apiutils.find_resource(
            share_client.share_types, parsed_args.share_type
        )

        if parsed_args.extra_specs:
            try:
                share_type.unset_keys(parsed_args.extra_specs)
            except Exception as e:
                raise exceptions.CommandError(
                    f"Failed to remove share type extra spec: {e}"
                )


class ListShareType(command.Lister):
    """List Share Types."""

    _description = _("List share types")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=_(
                'Display all share types whatever public or private. '
                'Default=False. (Admin only)'
            ),
        )
        parser.add_argument(
            '--extra-specs',
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_(
                'Filter share types with extra specs (key=value). '
                'Available only for API microversion >= 2.43. '
                'OPTIONAL: Default=None.'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        search_opts = {}
        if parsed_args.extra_specs:
            if share_client.api_version < api_versions.APIVersion("2.43"):
                raise exceptions.CommandError(
                    "Filtering by 'extra_specs' is available only with "
                    "API microversion '2.43' and above."
                )

            search_opts = {
                'extra_specs': utils.extract_extra_specs(
                    extra_specs={}, specs_to_add=parsed_args.extra_specs
                )
            }

        share_types = share_client.share_types.list(
            search_opts=search_opts, show_all=parsed_args.all
        )

        formatted_types = []
        for share_type in share_types:
            formatted_types.append(
                format_share_type(share_type, parsed_args.formatter)
            )

        values = (
            oscutils.get_dict_properties(s._info, ATTRIBUTES)
            for s in formatted_types
        )

        columns = utils.format_column_headers(ATTRIBUTES)

        return (columns, values)


class ShowShareType(command.ShowOne):
    """Show a share type."""

    _description = _("Display share type details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share_type',
            metavar="<share_type>",
            help=_("Share type to display (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_type = apiutils.find_resource(
            share_client.share_types, parsed_args.share_type
        )

        formatted_type = format_share_type(share_type, parsed_args.formatter)

        return (
            ATTRIBUTES,
            oscutils.get_dict_properties(formatted_type._info, ATTRIBUTES),
        )



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.osc.v2 import share_transfers as osc_share_transfers
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes

COLUMNS = [
    'ID',
    'Name',
    'Resource Type',
    'Resource Id',
    'Created At',
    'Source Project Id',
    'Destination Project Id',
    'Accepted',
    'Expires At',
]


class TestShareTransfer(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.shares_mock = self.app.client_manager.share.shares
        self.shares_mock.reset_mock()

        self.transfers_mock = self.app.client_manager.share.transfers
        self.transfers_mock.reset_mock()

        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )


class TestShareTransferCreate(TestShareTransfer):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.create.return_value = self.share

        self.shares_mock.get.return_value = self.share

        self.share_transfer = (
            manila_fakes.FakeShareTransfer.create_one_transfer()
        )
        self.transfers_mock.get.return_value = self.share_transfer
        self.transfers_mock.create.return_value = self.share_transfer

        self.cmd = osc_share_transfers.CreateShareTransfer(self.app, None)

        self.data = tuple(self.share_transfer._info.values())
        self.columns = tuple(self.share_transfer._info.keys())

    def test_share_transfer_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_transfer_create_required_args(self):
        arglist = [self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfers_mock.create.assert_called_with(self.share.id, name=None)

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareTransferDelete(TestShareTransfer):
    def setUp(self):
        super().setUp()

        self.transfer = manila_fakes.FakeShareTransfer.create_one_transfer()

        self.transfers_mock.get.return_value = self.transfer

        self.cmd = osc_share_transfers.DeleteShareTransfer(self.app, None)

    def test_share_transfer_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_transfer_delete(self):
        arglist = [self.transfer.id]
        verifylist = [('transfer', [self.transfer.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.transfers_mock.delete.assert_called_with(self.transfer.id)
        self.assertIsNone(result)

    def test_share_transfer_delete_multiple(self):
        transfers = manila_fakes.FakeShareTransfer.create_share_transfers(
            count=2
        )
        arglist = [transfers[0].id, transfers[1].id]
        verifylist = [('transfer', [transfers[0].id, transfers[1].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(self.transfers_mock.delete.call_count, len(transfers))
        self.assertIsNone(result)

    def test_share_transfer_delete_exception(self):
        arglist = [self.transfer.id]
        verifylist = [('transfer', [self.transfer.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.transfers_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareTransferShow(TestShareTransfer):
    def setUp(self):
        super().setUp()

        self.transfer = manila_fakes.FakeShareTransfer.create_one_transfer()
        self.transfers_mock.get.return_value = self.transfer

        self.cmd = osc_share_transfers.ShowShareTransfer(self.app, None)

        self.data = self.transfer._info.values()
        self.transfer._info.pop('auth_key')
        self.columns = self.transfer._info.keys()

    def test_share_transfer_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_transfer_show(self):
        arglist = [self.transfer.id]
        verifylist = [('transfer', self.transfer.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.transfers_mock.get.assert_called_with(self.transfer.id)
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareTransferList(TestShareTransfer):
    def setUp(self):
        super().setUp()

        self.transfers = manila_fakes.FakeShareTransfer.create_share_transfers(
            count=2
        )

        self.transfers_mock.list.return_value = self.transfers

        self.values = (
            oscutils.get_dict_properties(m._info, COLUMNS)
            for m in self.transfers
        )

        self.cmd = osc_share_transfers.ListShareTransfer(self.app, None)

    def test_list_transfers(self):
        arglist = ['--detailed']
        verifylist = [('detailed', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.transfers_mock.list.assert_called_with(
            detailed=1,
            search_opts={
                'all_tenants': False,
                'id': None,
                'name': None,
                'limit': None,
                'offset': None,
                'resource_type': None,
                'resource_id': None,
                'source_project_id': None,
            },
            sort_key=None,
            sort_dir=None,
        )

        self.assertEqual(COLUMNS, columns)
        self.assertEqual(list(self.values), list(data))


class TestShareTransferAccept(TestShareTransfer):
    def setUp(self):
        super().setUp()

        self.transfer = manila_fakes.FakeShareTransfer.create_one_transfer()

        self.transfers_mock.get.return_value = self.transfer

        self.cmd = osc_share_transfers.AcceptShareTransfer(self.app, None)

    def test_share_transfer_accept_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_transfer_accept(self):
        arglist = [self.transfer.id, self.transfer.auth_key]
        verifylist = [
            ('transfer', self.transfer.id),
            ('auth_key', self.transfer.auth_key),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.transfers_mock.accept.assert_called_with(
            self.transfer.id, self.transfer.auth_key, clear_access_rules=False
        )
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common import cliutils
from manilaclient.osc import utils as oscutils

LOG = logging.getLogger(__name__)


class CreateShareSnapshot(command.ShowOne):
    """Create a share snapshot."""

    _description = _("Create a snapshot of the given share")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share",
            metavar="<share>",
            help=_("Name or ID of the share to create snapshot of"),
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_(
                "Optional flag to indicate whether to snapshot "
                "a share even if it's busy. (Default=False)"
            ),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Add a name to the snapshot (Optional)."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Add a description to the snapshot (Optional)."),
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            default=False,
            help=_('Wait for share snapshot creation'),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this snapshot "
                "(repeat option to set multiple properties)."
                "Available only for microversion >= 2.73"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = utils.find_resource(share_client.shares, parsed_args.share)

        if share_client.api_version >= api_versions.APIVersion("2.73"):
            property = parsed_args.property or {}
        elif parsed_args.property:
            raise exceptions.CommandError(
                "Setting metadtaa is only available with manila API version "
                ">= 2.73"
            )

        share_snapshot = share_client.share_snapshots.create(
            share=share,
            force=parsed_args.force,
            name=parsed_args.name or None,
            description=parsed_args.description or None,
            metadata=property,
        )
        if parsed_args.wait:
            if not utils.wait_for_status(
                status_f=share_client.share_snapshots.get,
                res_id=share_snapshot.id,
                success_status=['available'],
            ):
                LOG.error(_("ERROR: Share snapshot is in error state."))

            share_snapshot = utils.find_resource(
                share_client.share_snapshots, share_snapshot.id
            )
        share_snapshot._info.pop('links', None)

        return self.dict2columns(share_snapshot._info)


class DeleteShareSnapshot(command.Command):
    """Delete one or more share snapshots"""

    _description = _("Delete one or more share snapshots")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            nargs="+",
            help=_("Name or ID of the snapshot(s) to delete"),
        )
        parser.add_argument(
            "--force",
            action='store_true',
            default=False,
            help=_("Delete the snapshot(s) ignoring the current state."),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share snapshot deletion"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for snapshot in parsed_args.snapshot:
            try:
                snapshot_obj = utils.find_resource(
                    share_client.share_snapshots, snapshot
                )
                if parsed_args.force:
                    share_client.share_snapshots.force_delete(snapshot_obj)
                else:
                    share_client.share_snapshots.delete(snapshot_obj)
                if parsed_args.wait:
                    if not utils.wait_for_delete(
                        manager=share_client.share_snapshots,
                        res_id=snapshot_obj.id,
                    ):
                        result += 1
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete snapshot with "
                        "name or ID '%(snapshot)s': %(e)s"
                    ),
                    {'snapshot': snapshot, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.snapshot)
            msg = _("%(result)s of %(total)s snapshots failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ShowShareSnapshot(command.ShowOne):
    """Display a share snapshot"""

    _description = _("Show details about a share snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Name or ID of the snapshot to display"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_snapshot = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        export_locations = share_client.share_snapshot_export_locations.list(
            share_snapshot
        )

        locations = []
        for location in export_locations:
            location._info.pop('links', None)
            locations.append(location._info)

        if parsed_args.formatter == 'table':
            locations = cliutils.convert_dict_list_to_string(locations)

        data = share_snapshot._info
        data['export_locations'] = locations
        # Special mapping for columns to make the output easier to read:
        # 'metadata' --> 'properties'
        data.update(
            {
                'properties': format_columns.DictColumn(
                    data.pop('metadata', {})
                ),
            },
        )
        data.pop('links', None)

        return self.dict2columns(data)


class SetShareSnapshot(command.Command):
    """Set share snapshot properties."""

    _description = _("Set share snapshot properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_('Name or ID of the snapshot to set a property for'),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Set a name to the snapshot."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Set a description to the snapshot."),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'manage_starting',
                'manage_error',
                'unmanage_starting',
                'unmanage_error',
                'error_deleting',
            ],
            help=_(
                "Assign a status to the snapshot (Admin only). "
                "Options include : available, error, creating, "
                "deleting, manage_starting, manage_error, "
                "unmanage_starting, unmanage_error, error_deleting."
            ),
        )
        parser.add_argument(
            "--property",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set a property to this snapshot "
                "(repeat option to set multiple properties)"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        share_snapshot = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        kwargs = {}

        if parsed_args.name is not None:
            kwargs['display_name'] = parsed_args.name
        if parsed_args.description is not None:
            kwargs['display_description'] = parsed_args.description

        try:
            share_client.share_snapshots.update(share_snapshot, **kwargs)
        except Exception as e:
            result += 1
            LOG.error(
                _(
                    "Failed to set share snapshot properties "
                    "'%(properties)s': %(exception)s"
                ),
                {'properties': kwargs, 'exception': e},
            )

        if parsed_args.status:
            try:
                share_client.share_snapshots.reset_state(
                    share_snapshot, parsed_args.status
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to update snapshot status to "
                        "'%(status)s': %(e)s"
                    ),
                    {'status': parsed_args.status, 'e': e},
                )
        if parsed_args.property:
            try:
                share_snapshot.set_metadata(parsed_args.property)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to set share snapshot properties "
                        "'%(properties)s': %(exception)s"
                    ),
                    {'properties': parsed_args.property, 'exception': e},
                )
                result += 1
        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class UnsetShareSnapshot(command.Command):
    """Unset a share snapshot property."""

    _description = _("Unset a share snapshot property")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Name or ID of the snapshot to set a property for"),
        )
        parser.add_argument(
            "--name",
            action='store_true',
            help=_("Unset snapshot name."),
        )
        parser.add_argument(
            "--description",
            action='store_true',
            help=_("Unset snapshot description."),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            help=_(
                'Remove a property from snapshot '
                '(repeat option to remove multiple properties)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_snapshot = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        kwargs = {}
        if parsed_args.name:
            kwargs['display_name'] = None
        if parsed_args.description:
            kwargs['display_description'] = None
        if kwargs:
            try:
                share_client.share_snapshots.update(share_snapshot, **kwargs)
            except Exception as e:
                msg = _(
                    "Failed to unset snapshot display name or display "
                    "description: %(e)s"
                )
                raise exceptions.CommandError(msg % {'e': e})
        if parsed_args.property:
            for key in parsed_args.property:
                try:
                    share_snapshot.delete_metadata([key])
                except Exception as e:
                    msg = _(
                        "Failed to unset snapshot property '%(key)s': %(e)s"
                    )
                    raise exceptions.CommandError(msg % {'key': key, 'e': e})


class ListShareSnapshot(command.Lister):
    """List snapshots."""

    _description = _("List snapshots")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--all-projects",
            action='store_true',
            default=False,
            help=_("Display snapshots from all projects (Admin only)."),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Filter results by name."),
        )
        parser.add_argument(
            '--description',
            metavar="<description>",
            default=None,
            help=_(
                "Filter results by description. Available only for "
                "microversion >= 2.36."
            ),
        )
        parser.add_argument(
            '--status',
            metavar="<status>",
            default=None,
            help=_('Filter results by status'),
        )
        parser.add_argument(
            '--share',
            metavar='<share>',
            default=None,
            help=_('Name or ID of a share to filter results by.'),
        )
        parser.add_argument(
            '--usage',
            metavar='<usage>',
            default=None,
            choices=['used', 'unused'],
            help=_("Option to filter snapshots by usage."),
        )
        parser.add_argument(
            "--limit",
            metavar="<num-snapshots>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of snapshots returned"),
        )
        parser.add_argument(
            "--marker",
            metavar="<snapshot>",
            help=_("The last share ID of the previous page"),
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            default='name:asc',
            help=_(
                "Sort output by selected keys and directions(asc or desc) "
                "(default: name:asc), multiple keys and directions can be "
                "specified separated by comma"
            ),
        )
        parser.add_argument(
            "--name~",
            metavar="<name~>",
            default=None,
            help=_(
                "Filter results matching a share snapshot name pattern. "
                "Available only for microversion >= 2.36."
            ),
        )
        parser.add_argument(
            '--description~',
            metavar="<description~>",
            default=None,
            help=_(
                "Filter results matching a share snapshot description "
                "pattern. Available only for microversion >= 2.36."
            ),
        )
        parser.add_argument(
            '--detail',
            action='store_true',
            default=False,
            help=_("List share snapshots with details"),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_(
                'Filter snapshots having a given metadata key=value '
                'property. (repeat option to filter by multiple '
                'properties)'
            ),
        )
        parser.add_argument(
            '--count',
            action='store_true',
            default=False,
            help=_(
                "The total count of share snapshots before pagination is "
                "applied. This parameter is useful when applying "
                "pagination parameters '--limit' and '--offset'. Available "
                "only for microversion >= 2.79."
            ),
        )

        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_id = None
        if parsed_args.share:
            share_id = utils.find_resource(
                share_client.shares, parsed_args.share
            ).id
        columns = ['ID', 'Name']

        search_opts = {
            'offset': parsed_args.marker,
            'limit': parsed_args.limit,
            'all_tenants': parsed_args.all_projects,
            'name': parsed_args.name,
            'status': parsed_args.status,
            'share_id': share_id,
            'usage': parsed_args.usage,
            'metadata': oscutils.extract_key_value_options(
                parsed_args.property
            ),
        }

        if share_client.api_version >= api_versions.APIVersion("2.36"):
            search_opts['name~'] = getattr(parsed_args, 'name~')
            search_opts['description~'] = getattr(parsed_args, 'description~')
            search_opts['description'] = parsed_args.description
        elif (
            parsed_args.description
            or getattr(parsed_args, 'name~')
            or getattr(parsed_args, 'description~')
        ):
            raise exceptions.CommandError(
                "Pattern based filtering (name~, description~ and description)"
                " is only available with manila API version >= 2.36"
            )

        if parsed_args.count:
            if share_client.api_version < api_versions.APIVersion("2.79"):
                raise exceptions.CommandError(
                    "Displaying total number of share snapshots is only "
                    "available with manila API version >= 2.79"
                )
            if parsed_args.formatter != 'table':
                raise exceptions.CommandError(
                    "Count can only be printed when using '--format table'"
                )

        if parsed_args.detail:
            columns.extend(
                [
                    'Status',
                    'Description',
                    'Created At',
                    'Size',
                    'Share ID',
                    'Share Proto',
                    'Share Size',
                    'User ID',
                ]
            )

        if parsed_args.all_projects:
            columns.append('Project ID')

        total_count = 0
        if parsed_args.count:
            search_opts['with_count'] = True
            snapshots, total_count = share_client.share_snapshots.list(
                search_opts=search_opts
            )
        else:
            snapshots = share_client.share_snapshots.list(
                search_opts=search_opts
            )

        snapshots = utils.sort_items(snapshots, parsed_args.sort, str)

        if parsed_args.count:
            print(f"Total number of snapshots: {total_count}")
        return (
            columns,
            (utils.get_item_properties(s, columns) for s in snapshots),
        )


class AdoptShareSnapshot(command.ShowOne):
    """Adopt a share snapshot not handled by Manila (Admin only)."""

    _description = _("Adopt a share snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share",
            metavar="<share>",
            help=_(
                "Name or ID of the share that owns the snapshot to be adopted."
            ),
        )
        parser.add_argument(
            "provider_location",
            metavar="<provider-location>",
            help=_("Provider location of the snapshot on the backend."),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Optional snapshot name (Default=None)."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Optional snapshot description (Default=None)."),
        )
        parser.add_argument(
            "--driver-option",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Set driver options as key=value pairs."
                "(repeat option to set multiple key=value pairs)"
            ),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            help=_("Wait until share snapshot is adopted"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = utils.find_resource(share_client.shares, parsed_args.share)

        snapshot = share_client.share_snapshots.manage(
            share=share,
            provider_location=parsed_args.provider_location,
            driver_options=parsed_args.driver_option,
            name=parsed_args.name,
            description=parsed_args.description,
        )

        if parsed_args.wait:
            if not utils.wait_for_status(
                status_f=share_client.share_snapshots.get,
                res_id=snapshot.id,
                success_status=['available'],
                error_status=['manage_error', 'error'],
            ):
                LOG.error(_("ERROR: Share snapshot is in error state."))

            snapshot = utils.find_resource(
                share_client.share_snapshots, snapshot.id
            )

        snapshot._info.pop('links', None)

        return self.dict2columns(snapshot._info)


class AbandonShareSnapshot(command.Command):
    """Abandon one or more share snapshots (Admin only)."""

    _description = _("Abandon share snapshot(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            nargs='+',
            help=_("Name or ID of the snapshot(s) to be abandoned."),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            help=_("Wait until share snapshot is abandoned"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for snapshot in parsed_args.snapshot:
            snapshot_obj = utils.find_resource(
                share_client.share_snapshots, snapshot
            )
            try:
                share_client.share_snapshots.unmanage(snapshot_obj)
                if parsed_args.wait:
                    if not utils.wait_for_delete(
                        manager=share_client.share_snapshots,
                        res_id=snapshot_obj.id,
                    ):
                        result += 1
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to abandon share snapshot with "
                        "name or ID '%(snapshot)s': %(e)s"
                    ),
                    {'snapshot': snapshot, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.snapshot)
            msg = _("%(result)s of %(total)s snapshots failed to abandon.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ShareSnapshotAccessAllow(command.ShowOne):
    """Allow read only access to a snapshot."""

    _description = _("Allow access to a snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Name or ID of the snapshot"),
        )
        parser.add_argument(
            'access_type',
            metavar="<access_type>",
            help=_(
                'Access rule type (only "ip", "user" (user or group), '
                '"cert" or "cephx" are supported).'
            ),
        )
        parser.add_argument(
            'access_to',
            metavar="<access_to>",
            help=_('Value that defines access.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        snapshot_obj = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        try:
            snapshot_access = share_client.share_snapshots.allow(
                snapshot=snapshot_obj,
                access_type=parsed_args.access_type,
                access_to=parsed_args.access_to,
            )
            return self.dict2columns(snapshot_access)

        except Exception as e:
            raise exceptions.CommandError(
                "Failed to create access to share snapshot "
                f"'{snapshot_obj}': {e}"
            )


class ShareSnapshotAccessDeny(command.Command):
    """Delete access to a snapshot"""

    _description = _("Delete access to a snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Name or ID of the share snapshot to deny access to."),
        )
        parser.add_argument(
            "id",
            metavar="<id>",
            nargs="+",
            help=_("ID(s) of the access rule(s) to be deleted."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        snapshot_obj = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        for access_id in parsed_args.id:
            try:
                share_client.share_snapshots.deny(
                    snapshot=snapshot_obj, id=access_id
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete access to share snapshot "
                        "for an access rule with ID '%(access)s': %(e)s"
                    ),
                    {'access': access_id, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.id)
            msg = _(
                "Failed to delete access to a share snapshot for "
                "%(result)s out of %(total)s access rule ID's "
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ShareSnapshotAccessList(command.Lister):
    """Show access list for a snapshot"""

    _description = _("Show access list for a snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_(
                "Name or ID of the share snapshot to show access list for."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        snapshot_obj = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        access_rules = share_client.share_snapshots.access_list(snapshot_obj)

        columns = ["ID", "Access Type", "Access To", "State"]

        return (
            columns,
            (utils.get_item_properties(s, columns) for s in access_rules),
        )


class ShareSnapshotListExportLocation(command.Lister):
    """List export locations of a given snapshot"""

    _description = _("List export locations of a given snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Name or ID of the share snapshot."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        snapshot_obj = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        export_locations = share_client.share_snapshot_export_locations.list(
            snapshot=snapshot_obj
        )

        columns = ["ID", "Path"]

        return (
            columns,
            (utils.get_item_properties(s, columns) for s in export_locations),
        )


class ShareSnapshotShowExportLocation(command.ShowOne):
    """Show export location of the share snapshot"""

    _description = _("Show export location of the share snapshot")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "snapshot",
            metavar="<snapshot>",
            help=_("Name or ID of the share snapshot."),
        )
        parser.add_argument(
            "export_location",
            metavar="<export-location>",
            help=_("ID of the share snapshot export location."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        snapshot_obj = utils.find_resource(
            share_client.share_snapshots, parsed_args.snapshot
        )

        export_location = share_client.share_snapshot_export_locations.get(
            export_location=parsed_args.export_location, snapshot=snapshot_obj
        )

        return self.dict2columns(export_location._info)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.api_versions import MAX_VERSION
from manilaclient.osc.v2 import share_backups as osc_share_backups
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes


class TestShareBackup(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.shares_mock = self.app.client_manager.share.shares
        self.shares_mock.reset_mock()

        self.backups_mock = self.app.client_manager.share.share_backups
        self.backups_mock.reset_mock()
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            MAX_VERSION
        )


class TestShareBackupCreate(TestShareBackup):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.get.return_value = self.share

        self.share_backup = manila_fakes.FakeShareBackup.create_one_backup(
            attrs={'status': 'available'}
        )
        self.backups_mock.create.return_value = self.share_backup
        self.backups_mock.get.return_value = self.share_backup
        self.cmd = osc_share_backups.CreateShareBackup(self.app, None)
        self.data = tuple(self.share_backup._info.values())
        self.columns = tuple(self.share_backup._info.keys())

    def test_share_backup_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_backup_create(self):
        arglist = [self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.backups_mock.create.assert_called_with(
            self.share,
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_backup_create_name(self):
        arglist = [self.share.id, '--name', "FAKE_SHARE_BACKUP_NAME"]
        verifylist = [
            ('share', self.share.id),
            ('name', "FAKE_SHARE_BACKUP_NAME"),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.backups_mock.create.assert_called_with(
            self.share,
            name="FAKE_SHARE_BACKUP_NAME",
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareBackupDelete(TestShareBackup):
    def setUp(self):
        super().setUp()

        self.share_backup = manila_fakes.FakeShareBackup.create_one_backup()
        self.backups_mock.get.return_value = self.share_backup

        self.cmd = osc_share_backups.DeleteShareBackup(self.app, None)

    def test_share_backup_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_backup_delete(self):
        arglist = [self.share_backup.id]
        verifylist = [('backup', [self.share_backup.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.backups_mock.delete.assert_called_with(self.share_backup)
        self.assertIsNone(result)

    def test_share_backup_delete_multiple(self):
        share_backups = manila_fakes.FakeShareBackup.create_share_backups(
            count=2
        )
        arglist = [share_backups[0].id, share_backups[1].id]
        verifylist = [('backup', [share_backups[0].id, share_backups[1].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertEqual(
            self.backups_mock.delete.call_count, len(share_backups)
        )
        self.assertIsNone(result)

    def test_share_backup_delete_exception(self):
        arglist = [self.share_backup.id]
        verifylist = [('backup', [self.share_backup.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.backups_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareBackupList(TestShareBackup):
    columns = [
        'ID',
        'Name',
        'Share ID',
        'Status',
    ]
    detailed_columns = [
        'ID',
        'Name',
        'Share ID',
        'Status',
        'Description',
        'Size',
        'Created At',
        'Updated At',
        'Availability Zone',
        'Progress',
        'Restore Progress',
        'Host',
        'Topic',
    ]

    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.get.return_value = self.share
        self.backups_list = manila_fakes.FakeShareBackup.create_share_backups(
            count=2
        )
        self.backups_mock.list.return_value = self.backups_list
        self.values = (
            oscutils.get_dict_properties(i._info, self.columns)
            for i in self.backups_list
        )
        self.detailed_values = (
            oscutils.get_dict_properties(i._info, self.detailed_columns)
            for i in self.backups_list
        )

        self.cmd = osc_share_backups.ListShareBackup(self.app, None)

    def test_share_backup_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.backups_mock.list.assert_called_with(
            detailed=0,
            search_opts={
                'offset': None,
                'limit': None,
                'name': None,
                'description': None,
                'name~': None,
                'description~': None,
                'status': None,
                'share_id': None,
            },
            sort_key=None,
            sort_dir=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_backup_list_detail(self):
        arglist = ['--detail']
        verifylist = [('detail', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.backups_mock.list.assert_called_with(
            detailed=1,
            search_opts={
                'offset': None,
                'limit': None,
                'name': None,
                'description': None,
                'name~': None,
                'description~': None,
                'status': None,
                'share_id': None,
            },
            sort_key=None,
            sort_dir=None,
        )
        self.assertEqual(self.detailed_columns, columns)
        self.assertEqual(list(self.detailed_values), list(data))

    def test_share_backup_list_for_share(self):
        arglist = ['--share', self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.backups_mock.list.assert_called_with(
            detailed=0,
            search_opts={
                'offset': None,
                'limit': None,
                'name': None,
                'description': None,
                'name~': None,
                'description~': None,
                'status': None,
                'share_id': self.share.id,
            },
            sort_key=None,
            sort_dir=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))


class TestShareBackupShow(TestShareBackup):
    def setUp(self):
        super().setUp()
        self.share_backup = manila_fakes.FakeShareBackup.create_one_backup()
        self.backups_mock.get.return_value = self.share_backup
        self.cmd = osc_share_backups.ShowShareBackup(self.app, None)
        self.data = tuple(self.share_backup._info.values())
        self.columns = tuple(self.share_backup._info.keys())

    def test_share_backup_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_backup_show(self):
        arglist = [self.share_backup.id]
        verifylist = [('backup', self.share_backup.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.backups_mock.get.assert_called_with(self.share_backup.id)

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareBackupRestore(TestShareBackup):
    def setUp(self):
        super().setUp()
        self.share_backup = manila_fakes.FakeShareBackup.create_one_backup()
        self.target_share = manila_fakes.FakeShare.create_one_share()
        self.backups_mock.get.return_value = self.share_backup
        self.shares_mock.get.return_value = self.target_share
        self.cmd = osc_share_backups.RestoreShareBackup(self.app, None)

    def test_share_backup_restore(self):
        arglist = [
            self.share_backup.id,
        ]
        verifylist = [
            ('backup', self.share_backup.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.backups_mock.restore.assert_called_with(self.share_backup.id)
        self.assertIsNone(result)

    def test_share_backup_restore_to_target(self):
        arglist = [
            self.share_backup.id,
            '--target-share',
            self.target_share.id,
        ]
        verifylist = [
            ('backup', self.share_backup.id),
            ('target_share', self.target_share.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.backups_mock.restore.assert_called_with(
            self.share_backup.id, target_share_id=self.target_share.id
        )
        self.assertIsNone(result)


class TestShareBackupSet(TestShareBackup):
    def setUp(self):
        super().setUp()
        self.share_backup = manila_fakes.FakeShareBackup.create_one_backup()
        self.backups_mock.get.return_value = self.share_backup
        self.cmd = osc_share_backups.SetShareBackup(self.app, None)

    def test_set_share_backup_name(self):
        arglist = [self.share_backup.id, '--name', "FAKE_SHARE_BACKUP_NAME"]
        verifylist = [
            ('backup', self.share_backup.id),
            ('name', "FAKE_SHARE_BACKUP_NAME"),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.backups_mock.update.assert_called_with(
            self.share_backup, name=parsed_args.name
        )
        self.assertIsNone(result)

    def test_set_backup_status(self):
        arglist = [self.share_backup.id, '--status', 'available']
        verifylist = [
            ('backup', self.share_backup.id),
            ('status', 'available'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.backups_mock.reset_status.assert_called_with(
            self.share_backup, parsed_args.status
        )
        self.assertIsNone(result)


class TestShareBackupUnset(TestShareBackup):
    def setUp(self):
        super().setUp()

        self.share_backup = manila_fakes.FakeShareBackup.create_one_backup()

        self.backups_mock.get.return_value = self.share_backup
        self.cmd = osc_share_backups.UnsetShareBackup(self.app, None)

    def test_unset_backup_name(self):
        arglist = [self.share_backup.id, '--name']
        verifylist = [('backup', self.share_backup.id), ('name', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.backups_mock.update.assert_called_with(
            self.share_backup, name=None
        )
        self.assertIsNone(result)

    def test_unset_backup_description(self):
        arglist = [self.share_backup.id, '--description']
        verifylist = [('backup', self.share_backup.id), ('description', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.backups_mock.update.assert_called_with(
            self.share_backup, description=None
        )
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from manilaclient import api_versions
from manilaclient.common._i18n import _


class QuotaSet(command.Command):
    """Set quotas for a project or project/user or project/share-type.

    It can be used to set the default class for all projects.
    """

    _description = _(
        "Set Quota for a project, or project/user or "
        "project/share-type or a class."
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        quota_type = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'project',
            metavar='<project/class>',
            help=_("A project (name/ID) or a class (e.g.: default)."),
        )
        quota_type.add_argument(
            '--class',
            dest='quota_class',
            action='store_true',
            default=False,
            help=_(
                "Update class quota to all projects. "
                "Mutually exclusive with '--user' and '--share-type'."
            ),
        )
        quota_type.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_(
                "Name or ID of a user to set the quotas for. "
                "Mutually exclusive with '--share-type' and '--class'."
            ),
        )
        quota_type.add_argument(
            '--share-type',
            metavar='<share-type>',
            type=str,
            default=None,
            help=_(
                "Name or ID of a share type to set the quotas for. "
                "Mutually exclusive with '--user' and '--class'. "
                "Available only for microversion >= 2.39"
            ),
        )
        parser.add_argument(
            '--shares',
            metavar='<shares>',
            type=int,
            default=None,
            help=_('New value for the "shares" quota.'),
        )
        parser.add_argument(
            '--snapshots',
            metavar='<snapshots>',
            type=int,
            default=None,
            help=_('New value for the "snapshots" quota.'),
        )
        parser.add_argument(
            '--gigabytes',
            metavar='<gigabytes>',
            type=int,
            default=None,
            help=_('New value for the "gigabytes" quota.'),
        )
        parser.add_argument(
            '--snapshot-gigabytes',
            metavar='<snapshot-gigabytes>',
            type=int,
            default=None,
            help=_('New value for the "snapshot-gigabytes" quota.'),
        )
        parser.add_argument(
            '--share-networks',
            metavar='<share-networks>',
            type=int,
            default=None,
            help=_('New value for the "share-networks" quota.'),
        )
        parser.add_argument(
            '--share-groups',
            metavar='<share-groups>',
            type=int,
            default=None,
            help=_(
                'New value for the "share-groups" quota. '
                'Available only for microversion >= 2.40'
            ),
        )
        parser.add_argument(
            '--share-group-snapshots',
            metavar='<share-group-snapshots>',
            type=int,
            default=None,
            help=_(
                'New value for the "share-group-snapshots" quota. '
                'Available only for microversion >= 2.40'
            ),
        )
        parser.add_argument(
            '--share-replicas',
            metavar='<share-replicas>',
            type=int,
            default=None,
            help=_(
                "Number of share replicas. "
                "Available only for microversion >= 2.53"
            ),
        )
        parser.add_argument(
            '--replica-gigabytes',
            metavar='<replica-gigabytes>',
            type=int,
            default=None,
            help=_(
                "Capacity of share replicas in total. "
                "Available only for microversion >= 2.53"
            ),
        )
        parser.add_argument(
            '--per-share-gigabytes',
            metavar='<per-share-gigabytes>',
            type=int,
            default=None,
            help=_(
                'New value for the "per-share-gigabytes" quota. '
                'Available only for microversion >= 2.62'
            ),
        )
        parser.add_argument(
            '--encryption-keys',
            metavar='<encryption-keys>',
            type=int,
            default=None,
            help=_(
                'New value for the "encryption-keys" quota. '
                'Available only for microversion >= 2.90'
            ),
        )
        parser.add_argument(
            '--force',
            dest='force',
            action="store_true",
            default=None,
            help=_('Force update the quota. Not applicable for class update.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        user_id = None
        if parsed_args.user:
            user_id = utils.find_resource(
                identity_client.users, parsed_args.user
            ).id

        kwargs = {
            "shares": parsed_args.shares,
            "snapshots": parsed_args.snapshots,
            "gigabytes": parsed_args.gigabytes,
            "snapshot_gigabytes": parsed_args.snapshot_gigabytes,
            "share_networks": parsed_args.share_networks,
        }

        if parsed_args.share_type is not None:
            if share_client.api_version < api_versions.APIVersion('2.39'):
                raise exceptions.CommandError(
                    _(
                        "'share type' quotas are available only starting with "
                        "'2.39' API microversion."
                    )
                )
            kwargs["share_type"] = parsed_args.share_type
        if parsed_args.share_groups is not None:
            if share_client.api_version < api_versions.APIVersion('2.40'):
                raise exceptions.CommandError(
                    _(
                        "'share group' quotas are available only "
                        "starting with '2.40' API microversion."
                    )
                )
            kwargs["share_groups"] = parsed_args.share_groups
        if parsed_args.share_group_snapshots is not None:
            if share_client.api_version < api_versions.APIVersion('2.40'):
                raise exceptions.CommandError(
                    _(
                        "'share group snapshots' quotas are available only "
                        "starting with '2.40' API microversion."
                    )
                )
            kwargs["share_group_snapshots"] = parsed_args.share_group_snapshots
        if parsed_args.share_replicas is not None:
            if share_client.api_version < api_versions.APIVersion('2.53'):
                raise exceptions.CommandError(
                    _(
                        "setting the number of 'share replicas' is available "
                        "only starting with API microversion '2.53'."
                    )
                )
            kwargs["share_replicas"] = parsed_args.share_replicas
        if parsed_args.replica_gigabytes is not None:
            if share_client.api_version < api_versions.APIVersion('2.53'):
                raise exceptions.CommandError(
                    _(
                        "setting the capacity of share replicas in total "
                        "is available only starting with API microversion "
                        "'2.53'."
                    )
                )
            kwargs["replica_gigabytes"] = parsed_args.replica_gigabytes
        if parsed_args.per_share_gigabytes is not None:
            if share_client.api_version < api_versions.APIVersion('2.62'):
                raise exceptions.CommandError(
                    _(
                        "'per share gigabytes' quotas are available only "
                        "starting with '2.62' API microversion."
                    )
                )
            kwargs["per_share_gigabytes"] = parsed_args.per_share_gigabytes
        if parsed_args.encryption_keys is not None:
            if share_client.api_version < api_versions.APIVersion('2.90'):
                raise exceptions.CommandError(
                    _(
                        "'encryption keys' quotas are available only "
                        "starting with '2.90' API microversion."
                    )
                )
            kwargs["encryption_keys"] = parsed_args.encryption_keys

        if all(value is None for value in kwargs.values()):
            raise exceptions.CommandError(
                _(
                    "Nothing to set. "
                    "New quota must be specified to at least one of the "
                    "following resources: 'shares', 'snapshots', 'gigabytes', "
                    "'snapshot-gigabytes', 'share-networks', 'share-type', "
                    "'share-groups', 'share-group-snapshots', "
                    "'share-replicas', 'replica-gigabytes', "
                    "'per-share-gigabytes', 'encryption_keys'"
                )
            )

        if parsed_args.quota_class:
            kwargs.update(
                {
                    "class_name": parsed_args.project,
                }
            )
            try:
                share_client.quota_classes.update(**kwargs)
            except Exception as e:
                msg = _(
                    "Failed to set quotas for class '%(project)s': '%(e)s'"
                )
                raise exceptions.CommandError(
                    msg % {'project': parsed_args.project, 'e': e}
                )
        else:
            project_id = utils.find_resource(
                identity_client.projects, parsed_args.project
            ).id

            kwargs.update(
                {
                    "tenant_id": project_id,
                    "force": parsed_args.force,
                    "user_id": user_id,
                }
            )

            try:
                share_client.quotas.update(**kwargs)
            except Exception as e:
                msg = _(
                    "Failed to set quotas for project '%(project)s': '%(e)s'"
                )
                raise exceptions.CommandError(
                    msg % {'project': parsed_args.project, 'e': e}
                )


class QuotaShow(command.ShowOne):
    """List the quotas for a project or project/user or project/share-type."""

    _description = _("Show Quota")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        quota_type = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Name or ID of the project to list quotas for.'),
        )
        quota_type.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_(
                "Name or ID of user to list the quotas for. Optional. "
                "Mutually exclusive with '--share-type'."
            ),
        )
        quota_type.add_argument(
            '--share-type',
            metavar='<share-type>',
            type=str,
            default=None,
            help=_(
                "Name or ID of a share type to list the quotas for. "
                "Optional. "
                "Mutually exclusive with '--user'. "
                "Available only for microversion >= 2.39"
            ),
        )
        parser.add_argument(
            '--detail',
            action='store_true',
            default=False,
            help=_(
                'Optional flag to indicate whether to show quota in detail.'
                ' Default false, available only for microversion >= 2.25.'
            ),
        )
        parser.add_argument(
            '--defaults',
            action='store_true',
            default=False,
            help=_('Show the default quotas for the project.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        user_id = None
        if parsed_args.user:
            user_id = utils.find_resource(
                identity_client.users, parsed_args.user
            ).id

        project_id = utils.find_resource(
            identity_client.projects, parsed_args.project
        ).id

        quotas = {}
        if parsed_args.defaults:
            quotas = share_client.quotas.defaults(project_id)
        else:
            kwargs = {
                "tenant_id": project_id,
                "user_id": user_id,
                "detail": parsed_args.detail,
            }
            if parsed_args.share_type is not None:
                if share_client.api_version < api_versions.APIVersion("2.39"):
                    raise exceptions.CommandError(
                        _(
                            "'share type' quotas are available only "
                            "starting with '2.39' API microversion."
                        )
                    )
                kwargs["share_type"] = parsed_args.share_type

            quotas = share_client.quotas.get(**kwargs)

        printable_quotas = {}
        for quota_k, quota_v in sorted(quotas.to_dict().items()):
            if isinstance(quota_v, dict):
                quota_v = '\n'.join(
                    [f'{k} = {v}' for k, v in sorted(quota_v.items())]
                )
            printable_quotas[quota_k] = quota_v

        return self.dict2columns(printable_quotas)


class QuotaDelete(command.Command):
    """Delete quota for project/user or project/share-type.

    The quota will revert back to default.
    """

    _description = _("Delete Quota")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        quota_type = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Name or ID of the project to delete quotas for.'),
        )
        quota_type.add_argument(
            '--user',
            metavar='<user>',
            default=None,
            help=_(
                "Name or ID of user to delete the quotas for. Optional. "
                "Mutually exclusive with '--share-type'."
            ),
        )
        quota_type.add_argument(
            '--share-type',
            metavar='<share-type>',
            type=str,
            default=None,
            help=_(
                "Name or ID of a share type to delete the quotas for. "
                "Optional. "
                "Mutually exclusive with '--user'. "
                "Available only for microversion >= 2.39"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        identity_client = self.app.client_manager.identity

        user_id = None
        if parsed_args.user:
            user_id = utils.find_resource(
                identity_client.users, parsed_args.user
            ).id

        project_id = utils.find_resource(
            identity_client.projects, parsed_args.project
        ).id

        kwargs = {"tenant_id": project_id, "user_id": user_id}
        if parsed_args.share_type:
            if share_client.api_version < api_versions.APIVersion("2.39"):
                raise exceptions.CommandError(
                    _(
                        "'share type' quotas are available only starting with "
                        "API microversion '2.39'."
                    )
                )
            kwargs["share_type"] = parsed_args.share_type

        share_client.quotas.delete(**kwargs)



# Copyright 2023 Cloudification GmbH.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common import constants
from manilaclient.osc import utils


LOG = logging.getLogger(__name__)


class CreateShareBackup(command.ShowOne):
    """Create a share backup."""

    _description = _("Create a backup of the given share")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "share",
            metavar="<share>",
            help=_("Name or ID of the share to backup."),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            default=None,
            help=_('Optional share backup name. (Default=None).'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default=None,
            help=_('Optional share backup description. (Default=None).'),
        )
        parser.add_argument(
            "--backup-options",
            metavar="<key=value>",
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                "Backup driver option key=value pairs (Optional, "
                "Default=None)."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share = osc_utils.find_resource(share_client.shares, parsed_args.share)

        body = {}
        if parsed_args.backup_options:
            body['backup_options'] = utils.extract_key_value_options(
                parsed_args.backup_options
            )
        if parsed_args.description:
            body['description'] = parsed_args.description
        if parsed_args.name:
            body['name'] = parsed_args.name

        share_backup = share_client.share_backups.create(share, **body)
        share_backup._info.pop('links', None)
        return self.dict2columns(share_backup._info)


class DeleteShareBackup(command.Command):
    """Delete one or more share backups."""

    _description = _("Delete one or more share backups")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            nargs="+",
            help=_("Name or ID of the backup(s) to delete"),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share backup deletion"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for backup in parsed_args.backup:
            try:
                share_backup_obj = osc_utils.find_resource(
                    share_client.share_backups, backup
                )
                share_client.share_backups.delete(share_backup_obj)

                if parsed_args.wait:
                    if not osc_utils.wait_for_delete(
                        manager=share_client.share_backups,
                        res_id=share_backup_obj.id,
                    ):
                        result += 1

            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete a share backup with "
                        "name or ID '%(backup)s': %(e)s"
                    ),
                    {'backup': backup, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.backup)
            msg = _("%(result)s of %(total)s backups failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListShareBackup(command.Lister):
    """List share backups."""

    _description = _("List share backups")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--share",
            metavar="<share>",
            default=None,
            help=_("Name or ID of the share to list backups for."),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Filter results by name. Default=None."),
        )
        parser.add_argument(
            '--description',
            metavar="<description>",
            default=None,
            help=_("Filter results by description. Default=None."),
        )
        parser.add_argument(
            "--name~",
            metavar="<name~>",
            default=None,
            help=_("Filter results matching a share backup name pattern. "),
        )
        parser.add_argument(
            '--description~',
            metavar="<description~>",
            default=None,
            help=_("Filter results matching a share backup description "),
        )
        parser.add_argument(
            '--status',
            metavar="<status>",
            default=None,
            help=_('Filter results by status. Default=None.'),
        )
        parser.add_argument(
            "--limit",
            metavar="<num-backups>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of backups returned. Default=None."),
        )
        parser.add_argument(
            '--offset',
            metavar="<offset>",
            default=None,
            help='Start position of backup records listing.',
        )
        parser.add_argument(
            '--sort-key',
            '--sort_key',
            metavar='<sort_key>',
            type=str,
            default=None,
            help=(
                f'Key to be sorted, available keys are '
                f'{constants.BACKUP_SORT_KEY_VALUES}. Default=None.'
            ),
        )
        parser.add_argument(
            '--sort-dir',
            '--sort_dir',
            metavar='<sort_dir>',
            type=str,
            default=None,
            help=(
                f'Sort direction, available values are '
                f'{constants.SORT_DIR_VALUES}. OPTIONAL: Default=None.'
            ),
        )
        parser.add_argument(
            '--detail',
            dest='detail',
            metavar='<0|1>',
            nargs='?',
            type=int,
            const=1,
            default=0,
            help="Show detailed information about share backups.",
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share_id = None
        if parsed_args.share:
            share_id = osc_utils.find_resource(
                share_client.shares, parsed_args.share
            ).id
        columns = ['ID', 'Name', 'Share ID', 'Status']

        if parsed_args.detail:
            columns.extend(
                [
                    'Description',
                    'Size',
                    'Created At',
                    'Updated At',
                    'Availability Zone',
                    'Progress',
                    'Restore Progress',
                    'Host',
                    'Topic',
                ]
            )

        search_opts = {
            'limit': parsed_args.limit,
            'offset': parsed_args.offset,
            'name': parsed_args.name,
            'description': parsed_args.description,
            'status': parsed_args.status,
            'share_id': share_id,
        }

        search_opts['name~'] = getattr(parsed_args, 'name~')
        search_opts['description~'] = getattr(parsed_args, 'description~')

        backups = share_client.share_backups.list(
            detailed=parsed_args.detail,
            search_opts=search_opts,
            sort_key=parsed_args.sort_key,
            sort_dir=parsed_args.sort_dir,
        )

        return (
            columns,
            (osc_utils.get_item_properties(b, columns) for b in backups),
        )


class ShowShareBackup(command.ShowOne):
    """Show share backup."""

    _description = _("Show details of a backup")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup", metavar="<backup>", help=_("ID of the share backup. ")
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        backup = osc_utils.find_resource(
            share_client.share_backups, parsed_args.backup
        )
        backup._info.pop('links', None)
        return self.dict2columns(backup._info)


class RestoreShareBackup(command.Command):
    """Restore share backup to share"""

    _description = _("Attempt to restore share backup")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup", metavar="<backup>", help=_('ID of backup to restore.')
        )
        parser.add_argument(
            "--target-share",
            metavar="<target-share>",
            default=None,
            help=_(
                'share to restore backup to. Source share if none supplied'
            ),
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            default=False,
            help=_('Wait for restore conclusion'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        kwargs = {}

        share_backup = osc_utils.find_resource(
            share_client.share_backups, parsed_args.backup
        )
        target_share_id = None
        if parsed_args.target_share is not None:
            if share_client.api_version < api_versions.APIVersion('2.91'):
                raise exceptions.CommandError(
                    'performing targeted restores is only available '
                    'for API microversion >= 2.91'
                )
            else:
                target_share_id = osc_utils.find_resource(
                    share_client.shares, parsed_args.target_share
                ).id
                kwargs['target_share_id'] = target_share_id

        share_client.share_backups.restore(share_backup.id, **kwargs)

        if parsed_args.wait:
            if not osc_utils.wait_for_status(
                status_f=share_client.shares.get,
                res_id=(target_share_id or share_backup.share_id),
                success_status=['available'],
                error_status=['error', 'backup_restoring_error'],
            ):
                LOG.error(_("ERROR: share is in error state."))


class SetShareBackup(command.Command):
    """Set share backup properties."""

    _description = _("Set share backup properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            help=_('Name or ID of the backup to set a property for'),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            default=None,
            help=_("Set a name to the backup."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("Set a description to the backup."),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=[
                'available',
                'error',
                'creating',
                'deleting',
                'restoring',
            ],
            help=_(
                "Assign a status to the backup(Admin only). "
                "Options include : available, error, creating, "
                "deleting, restoring."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        share_backup = osc_utils.find_resource(
            share_client.share_backups, parsed_args.backup
        )

        kwargs = {}

        if parsed_args.name is not None:
            kwargs['name'] = parsed_args.name
        if parsed_args.description is not None:
            kwargs['description'] = parsed_args.description

        try:
            share_client.share_backups.update(share_backup, **kwargs)
        except Exception as e:
            result += 1
            LOG.error(
                _(
                    "Failed to set share backup properties "
                    "'%(properties)s': %(exception)s"
                ),
                {'properties': kwargs, 'exception': e},
            )

        if parsed_args.status:
            try:
                share_client.share_backups.reset_status(
                    share_backup, parsed_args.status
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _("Failed to update backup status to '%(status)s': %(e)s"),
                    {'status': parsed_args.status, 'e': e},
                )
        if result > 0:
            raise exceptions.CommandError(
                _("One or more of the set operations failed")
            )


class UnsetShareBackup(command.Command):
    """Unset share backup properties."""

    _description = _("Unset share backup properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            help=_('Name or ID of the backup to unset a property for'),
        )
        parser.add_argument(
            "--name",
            action='store_true',
            help=_("Unset a name to the backup."),
        )
        parser.add_argument(
            "--description",
            action='store_true',
            help=_("Unset a description to the backup."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        share_backup = osc_utils.find_resource(
            share_client.share_backups, parsed_args.backup
        )

        kwargs = {}
        if parsed_args.name:
            kwargs['name'] = None
        if parsed_args.description:
            kwargs['description'] = None
        if not kwargs:
            msg = "Either name or description must be provided."
            raise exceptions.CommandError(msg)

        try:
            share_client.share_backups.update(share_backup, **kwargs)
        except Exception as e:
            LOG.error(
                _(
                    "Failed to unset share backup properties "
                    "'%(properties)s': %(exception)s"
                ),
                {'properties': kwargs, 'exception': e},
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json
import time

from tempest.lib.cli import base
from tempest.lib.common.utils import data_utils

from manilaclient import config

CONF = config.CONF


class OSCClientTestBase(base.ClientTestBase):
    """Base class for OSC manila functional tests"""

    @classmethod
    def get_admin_client(cls):
        admin_client = base.CLIClient(
            username=CONF.admin_username,
            password=CONF.admin_password,
            tenant_name=CONF.admin_tenant_name,
            uri=CONF.admin_auth_url,
            cli_dir=CONF.manila_exec_dir,
            insecure=CONF.insecure,
            project_domain_name=CONF.admin_project_domain_name or None,
            project_domain_id=CONF.admin_project_domain_id or None,
            user_domain_name=CONF.admin_user_domain_name or None,
            user_domain_id=CONF.admin_user_domain_id or None,
        )
        return admin_client

    @classmethod
    def get_user_client(cls):
        user_client = base.CLIClient(
            username=CONF.username,
            password=CONF.password,
            tenant_name=CONF.tenant_name,
            uri=CONF.auth_url,
            cli_dir=CONF.manila_exec_dir,
            insecure=CONF.insecure,
            project_domain_name=CONF.project_domain_name or None,
            project_domain_id=CONF.project_domain_id or None,
            user_domain_name=CONF.user_domain_name or None,
            user_domain_id=CONF.user_domain_id or None,
        )
        return user_client

    @property
    def admin_client(self):
        if not hasattr(self, '_admin_client'):
            self._admin_client = self.get_admin_client()
        return self._admin_client

    @property
    def user_client(self):
        if not hasattr(self, '_user_client'):
            self._user_client = self.get_user_client()
        return self._user_client

    def _get_clients(self):
        return self.admin_client

    def _get_property_from_output(self, output):
        """Creates a dictionary from the given output"""
        obj = {}
        items = self.parser.listing(output)
        for item in items:
            obj[item['Field']] = str(item['Value'])
        return obj

    def _wait_for_object_status(
        self,
        object_name,
        object_id,
        status,
        timeout=CONF.build_timeout,
        interval=CONF.build_interval,
    ):
        """Waits for a object to reach a given status."""

        start_time = time.time()
        while time.time() - start_time < timeout:
            if (
                status
                == self.openstack(
                    f'{object_name} show -c status -f value {object_id}'
                ).rstrip()
            ):
                break
            time.sleep(interval)
        else:
            self.fail(
                f"{object_name} {object_id} did not reach status {status} "
                f"after {timeout} seconds."
            )

    def check_object_deleted(
        self,
        object_name,
        object_id,
        timeout=CONF.build_timeout,
        interval=CONF.build_interval,
    ):
        """Check that object deleted successfully"""

        cmd = f'{object_name} list -c ID -f value'
        start_time = time.time()
        while time.time() - start_time < timeout:
            if object_id not in self.openstack(cmd):
                break
            time.sleep(interval)
        else:
            self.fail(
                f"{object_name} {object_id} not deleted after {timeout} "
                f"seconds."
            )

    def openstack(
        self,
        action,
        flags='',
        params='',
        fail_ok=False,
        merge_stderr=False,
        client=None,
    ):
        """Executes openstack command for given action"""

        if '--os-share-api-version' not in flags:
            flags = (
                flags + f'--os-share-api-version {CONF.max_api_microversion}'
            )

        if client is None:
            client = self.admin_client

        return client.openstack(
            action,
            flags=flags,
            params=params,
            fail_ok=fail_ok,
            merge_stderr=merge_stderr,
        )

    def listing_result(self, object_name, command, client=None):
        """Returns output for the given command as list of dictionaries"""

        output = self.openstack(object_name, params=command, client=client)
        result = self.parser.listing(output)
        return result

    def dict_result(self, object_name, command, client=None):
        """Returns output for the given command as dictionary"""

        output = self.openstack(object_name, params=command, client=client)
        result_dict = self._get_property_from_output(output)
        return result_dict

    def create_share(
        self,
        share_protocol=None,
        size=None,
        name=None,
        snapshot_id=None,
        properties=None,
        share_network=None,
        description=None,
        public=False,
        share_type=None,
        availability_zone=None,
        share_group=None,
        add_cleanup=True,
        client=None,
        wait=None,
        wait_for_status='available',
    ):
        name = name or data_utils.rand_name('autotest_share_name')
        # share_type = dhss_false until we have implemented
        # share network commands for osc
        share_type = share_type or 'dhss_false'

        cmd = 'create {protocol} {size} {name} {desc} {public} {stype}'.format(
            protocol=share_protocol or 'NFS',
            size=size or '1',
            name=f'--name {name}',
            desc=f'--description {description}',
            public=f'--public {public}',
            stype=f'--share-type {share_type}',
        )

        if snapshot_id:
            cmd = cmd + f' --snapshot-id {snapshot_id}'
        if properties:
            for key, value in properties.items():
                cmd = cmd + f' --property {key}={value}'
        if share_network:
            cmd = cmd + f' --share-network {share_network}'
        if availability_zone:
            cmd = cmd + f' --availability-zone {availability_zone}'
        if share_group:
            cmd = cmd + f' --share-group {share_group}'
        if wait:
            cmd = cmd + ' --wait'

        share_object = self.dict_result('share', cmd, client=client)
        self._wait_for_object_status(
            'share', share_object['id'], wait_for_status
        )

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                'share delete {} --wait'.format(share_object['id']),
            )
        return share_object

    def list_pools(self, backend=None, host=None, pool=None, detail=False):
        cmd = 'pool list '

        if backend:
            cmd += f'--backend {backend} '
        if pool:
            cmd += f'--pool {pool} '
        if host:
            cmd += f'--host {host} '
        if detail:
            cmd += '--detail'

        pools = self.listing_result('share', cmd)

        return pools

    def create_share_type(
        self,
        name=None,
        dhss=False,
        description=None,
        snapshot_support=None,
        create_share_from_snapshot_support=None,
        revert_to_snapshot_support=False,
        mount_snapshot_support=False,
        extra_specs={},
        public=True,
        add_cleanup=True,
        client=None,
        formatter=None,
    ):
        name = name or data_utils.rand_name('autotest_share_type_name')

        cmd = f'create {name} {dhss} --public {public}'
        if description:
            cmd += f' --description {description}'
        if snapshot_support:
            cmd += f' --snapshot-support {snapshot_support}'
        if create_share_from_snapshot_support:
            cmd += (
                ' --create-share-from-snapshot-support '
                f'{create_share_from_snapshot_support}'
            )
        if revert_to_snapshot_support:
            cmd += (
                f' --revert-to-snapshot-support  {revert_to_snapshot_support}'
            )
        if mount_snapshot_support:
            cmd += f' --mount-snapshot-support {mount_snapshot_support}'
        if extra_specs:
            specs = ' --extra-specs'
            for key, value in extra_specs.items():
                specs += f' {key}={value}'
            cmd += specs

        if formatter == 'json':
            cmd = f'share type {cmd} -f {formatter} '
            share_type = json.loads(self.openstack(cmd, client=client))
        else:
            share_type = self.dict_result('share type', cmd, client=client)

        if add_cleanup:
            self.addCleanup(
                self.openstack, f'share type delete {share_type["id"]}'
            )
        return share_type

    def list_services(self, host=None, status=None, state=None, zone=None):
        cmd = 'service list '

        if host:
            cmd += f'--host {host} '
        if status:
            cmd += f'--status {status} '
        if state:
            cmd += f'--state {state} '
        if zone:
            cmd += f'--zone {zone} '

        services = self.listing_result('share', cmd)
        return services

    def create_share_access_rule(
        self,
        share,
        access_type,
        access_to,
        properties=None,
        access_level=None,
        wait=False,
        lock_visibility=False,
        lock_deletion=False,
        lock_reason=None,
        add_cleanup=False,
    ):
        cmd = f'access create {share} {access_type} {access_to} '

        if access_level:
            cmd += f'--access-level {access_level} '
        if properties:
            cmd += f'--properties {properties} '
        if wait:
            cmd += '--wait '
        if lock_visibility:
            cmd += '--lock-visibility '
        if lock_deletion:
            cmd += '--lock-deletion '
        if lock_reason:
            cmd += f'--lock-reason {lock_reason}'

        access_rule = self.dict_result('share', cmd)

        return access_rule

    def get_share_export_locations(self, share):
        cmd = f'export location list {share}'
        export_locations = json.loads(self.openstack(f'share {cmd} -f json'))
        return export_locations

    def create_snapshot(
        self,
        share,
        name=None,
        description=None,
        wait=True,
        force=None,
        add_cleanup=True,
        client=None,
    ):
        name = name or data_utils.rand_name('autotest_snapshot_name')

        cmd = f'snapshot create {share} --name {name} '

        if description:
            cmd += f' --description {description}'
        if wait:
            cmd += ' --wait'
        if force:
            cmd += ' --force'

        snapshot_object = self.dict_result('share', cmd, client=client)

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                f'share snapshot delete {snapshot_object["id"]} --wait',
            )

        return snapshot_object

    def create_share_transfer(self, share, name=None, client=None):
        name = name or data_utils.rand_name('autotest_share_transfer_name')
        cmd = f'transfer create {share} --name {name} '
        transfer_object = self.dict_result('share', cmd, client=client)

        return transfer_object

    def create_share_network(
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        name=None,
        description=None,
        availability_zone=None,
        add_cleanup=True,
    ):
        name = name or data_utils.rand_name('autotest_share_network_name')
        cmd = f'network create --name {name} --description {description}'
        if neutron_net_id:
            cmd = cmd + f' --neutron-net-id {neutron_net_id}'
        if neutron_subnet_id:
            cmd = cmd + f' --neutron-subnet-id {neutron_subnet_id}'
        if availability_zone:
            cmd = cmd + f' --availability-zone {availability_zone}'

        share_network_obj = self.dict_result('share', cmd)
        self._wait_for_object_status(
            'share network', share_network_obj['id'], 'active'
        )
        if add_cleanup:
            self.addCleanup(
                self.openstack,
                f'share network delete {share_network_obj["id"]}',
            )
        return share_network_obj

    def create_share_replica(
        self,
        share,
        availability_zone=None,
        share_network=None,
        wait=None,
        properties=None,
        add_cleanup=True,
    ):
        cmd = f'replica create {share}'

        if availability_zone:
            cmd = cmd + f' --availability-zone {availability_zone}'
        if wait:
            cmd = cmd + ' --wait'
        if share_network:
            cmd = cmd + f' --share-network {share_network}'
        if properties:
            for key, value in properties.items():
                cmd = cmd + f' --property {key}={value}'

        replica_object = self.dict_result('share', cmd)
        self._wait_for_object_status(
            'share replica', replica_object['id'], 'available'
        )

        if add_cleanup:
            self.addCleanup(
                self._delete_share_replica_and_wait, replica_object['id']
            )
        return replica_object

    def _delete_share_replica_and_wait(self, replica_id):
        """Delete a share replica and wait for it to be gone.

        Uses check_object_deleted (respects CONF.build_timeout) instead of
        the CLI --wait flag which is limited to osc_lib's hardcoded 300s.
        """
        self.openstack(
            f'share replica delete {replica_id}',
            fail_ok=True,
        )
        self.check_object_deleted('share replica', replica_id)

    def get_share_replica_export_locations(self, replica):
        cmd = f'replica export location list {replica}'
        export_locations = self.listing_result('share', cmd)
        return export_locations

    def create_share_group_type(
        self,
        name=None,
        share_types=None,
        group_specs=None,
        public=True,
        add_cleanup=True,
    ):
        name = name or data_utils.rand_name('autotest_share_group_types_name')
        share_types = share_types or 'None'

        cmd = f'group type create {name} {share_types} '

        if group_specs:
            cmd = cmd + f' --group-specs {group_specs} '
        if not public:
            cmd = cmd + f' --public {public} '

        share_object = self.dict_result('share', cmd)

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                'share group type delete {}'.format(share_object['id']),
            )
        return share_object

    def share_group_type_access_create(self, group_type, project):
        cmd = f'group type access create {group_type} {project} '

        self.dict_result('share', cmd)

    def share_group_type_access_delete(self, group_type, access_id):
        cmd = f'group type access delete {group_type} {access_id} '

        self.dict_result('share', cmd)

    def check_create_network_subnet(
        self,
        share_network,
        neutron_net_id=None,
        neutron_subnet_id=None,
        availability_zone=None,
        restart_check=None,
    ):
        cmd = f'network subnet create {share_network} --check-only'

        if neutron_net_id:
            cmd += f' --neutron-net-id {neutron_net_id}'
        if neutron_subnet_id:
            cmd += f' --neutron-subnet-id {neutron_subnet_id}'
        if availability_zone:
            cmd += f' --availability-zone {availability_zone}'
        if restart_check:
            cmd += ' --restart-check'

        check_result = self.dict_result('share', cmd)
        return check_result

    def create_resource_lock(
        self,
        resource_id,
        resource_type='share',
        resource_action='delete',
        lock_reason=None,
        add_cleanup=True,
        client=None,
    ):
        cmd = f'lock create {resource_id} {resource_type}'
        cmd += f' --resource-action {resource_action}'

        if lock_reason:
            cmd += f' --reason "{lock_reason}"'

        lock = self.dict_result('share', cmd, client=client)

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                'share lock delete {}'.format(lock['id']),
                client=client,
            )
        return lock

    def create_backup(
        self,
        share_id,
        name=None,
        description=None,
        backup_options=None,
        add_cleanup=True,
    ):
        name = name or data_utils.rand_name('autotest_backup_name')

        cmd = f'backup create {share_id} '

        if name:
            cmd += f' --name {name}'
        if description:
            cmd += f' --description {description}'
        if backup_options:
            options = ' --backup-options'
            for key, value in backup_options.items():
                options += f' {key}={value}'
            cmd += options

        backup_object = self.dict_result('share', cmd)
        self._wait_for_object_status(
            'share backup', backup_object['id'], 'available'
        )

        if add_cleanup:
            self.addCleanup(
                self.openstack,
                f'share backup delete {backup_object["id"]} --wait',
            )

        return backup_object

    def create_qos_type(
        self,
        name=None,
        description=None,
        specs=None,
        add_cleanup=True,
        client=None,
        formatter=None,
    ):
        name = name or data_utils.rand_name('autotest_qos_type_name')
        specs = specs or {}

        cmd = f'create {name} '
        if description:
            cmd += f' --description {description}'
        if specs:
            q_specs = ''
            for key, value in specs.items():
                q_specs += f' --spec {key}={value}'
            cmd += q_specs

        if formatter == 'json':
            cmd = f'share qos type {cmd} -f {formatter} '
            qos_type = json.loads(self.openstack(cmd, client=client))
        else:
            qos_type = self.dict_result('share qos type', cmd, client=client)

        if add_cleanup:
            self.addCleanup(
                self.openstack, f'share qos type delete {qos_type["id"]}'
            )
        return qos_type



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from manilaclient.tests.functional.osc import base


class ShareNetworksCLITest(base.OSCClientTestBase):
    def test_openstack_share_network_create(self):
        share_network_name = 'test_create_share_network'
        share_network = self.create_share_network(name=share_network_name)

        self.assertEqual(share_network['name'], share_network_name)

        share_network_list = self.listing_result('share network', 'list')
        self.assertIn(
            share_network['id'], [item['ID'] for item in share_network_list]
        )

    def test_openstack_share_network_list(self):
        share_network = self.create_share_network()
        share_network_list = self.listing_result('share network', 'list')
        self.assertTableStruct(
            share_network_list,
            [
                'ID',
                'Name',
            ],
        )
        self.assertIn(
            share_network['id'], [item['ID'] for item in share_network_list]
        )

    def test_openstack_share_network_show(self):
        share_network = self.create_share_network()

        result = self.dict_result(
            'share network', 'show {}'.format(share_network['id'])
        )
        self.assertEqual(share_network['id'], result['id'])

        listing_result = self.listing_result(
            'share network', 'show {}'.format(share_network['id'])
        )
        self.assertTableStruct(listing_result, ['Field', 'Value'])

    def test_openstack_share_network_delete(self):
        share_network = self.create_share_network(add_cleanup=False)
        share_network_list = self.listing_result('share network', 'list')

        self.assertIn(
            share_network['id'], [item['ID'] for item in share_network_list]
        )

        self.openstack('share network delete {}'.format(share_network['id']))
        self.check_object_deleted('share network', share_network['id'])
        share_network_list_after_delete = self.listing_result(
            'share network', 'list'
        )

        self.assertNotIn(
            share_network['id'],
            [item['ID'] for item in share_network_list_after_delete],
        )

    def test_openstack_share_network_set(self):
        share_network = self.create_share_network()
        self.openstack(
            'share network set {} --name {}'.format(
                share_network['id'], 'new_name'
            )
        )
        result = self.dict_result(
            'share network', 'show {}'.format(share_network['id'])
        )
        self.assertEqual(share_network['id'], result['id'])
        self.assertEqual('new_name', result['name'])

    def test_openstack_share_network_unset(self):
        share_network = self.create_share_network(name='test_name')
        result1 = self.dict_result(
            'share network', 'show {}'.format(share_network['id'])
        )
        self.assertEqual(share_network['id'], result1['id'])
        self.assertEqual(share_network['name'], result1['name'])

        self.openstack(
            'share network unset {} --name'.format(share_network['id'])
        )
        result2 = self.dict_result(
            'share network', 'show {}'.format(share_network['id'])
        )
        self.assertEqual(share_network['id'], result2['id'])
        self.assertEqual('None', result2['name'])



# Copyright 2019 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base

RESOURCES_PATH = '/share-networks/%(share_network_id)s/subnets'
RESOURCE_PATH = RESOURCES_PATH + '/%(share_network_subnet_id)s'
RESOURCE_NAME = 'share_network_subnet'


class ShareNetworkSubnet(base.MetadataCapableResource):
    """Network subnet info for Manila share networks."""

    def __repr__(self):
        return f"<ShareNetworkSubnet: {self.id}>"

    def __getitem__(self, key):
        return self._info[key]

    def delete(self):
        """Delete this share network subnet."""
        self.manager.delete(self)


class ShareNetworkSubnetManager(base.MetadataCapableManager):
    """Manage :class:`ShareNetworkSubnet` resources."""

    resource_class = ShareNetworkSubnet
    resource_path = '/share-networks'
    subresource_path = '/subnets'

    def _do_create(
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        availability_zone=None,
        share_network_id=None,
        metadata=None,
    ):
        """Create share network subnet.

        :param neutron_net_id: ID of Neutron network
        :param neutron_subnet_id: ID of Neutron subnet
        :param availability_zone: Name of the target availability zone
        :param metadata: dict - optional metadata to set on share creation
        :rtype: :class:`ShareNetworkSubnet`
        """
        values = {}
        if neutron_net_id:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id:
            values['neutron_subnet_id'] = neutron_subnet_id
        if availability_zone:
            values['availability_zone'] = availability_zone
        if metadata:
            values['metadata'] = metadata

        body = {'share-network-subnet': values}
        url = f'/share-networks/{share_network_id}/subnets'

        return self._create(url, body, RESOURCE_NAME)

    @api_versions.wraps("2.0", "2.77")
    def create(
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        availability_zone=None,
        share_network_id=None,
    ):
        return self._do_create(
            neutron_net_id,
            neutron_subnet_id,
            availability_zone,
            share_network_id,
        )

    @api_versions.wraps("2.78")
    def create(  # noqa
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        availability_zone=None,
        share_network_id=None,
        metadata=None,
    ):
        return self._do_create(
            neutron_net_id,
            neutron_subnet_id,
            availability_zone,
            share_network_id,
            metadata,
        )

    def get(self, share_network, share_network_subnet):
        """Get a share network subnet.

        :param share_network_subnet: share network subnet to get.
        :rtype: :class:`ShareNetworkSubnet`
        """
        share_network_id = base.getid(share_network)
        share_network_subnet_id = base.getid(share_network_subnet)
        url = (
            f'/share-networks/{share_network_id}/subnets'
            f'/{share_network_subnet_id}'
        )
        return self._get(url, "share_network_subnet")

    def delete(self, share_network, share_network_subnet):
        """Delete a share network subnet.

        :param share_network: share network that owns the subnet.
        :param share_network_subnet: share network subnet to be deleted.
        """
        url = (
            f'/share-networks/{base.getid(share_network)}/subnets'
            f'/{share_network_subnet}'
        )
        self._delete(url)

    @api_versions.wraps('2.78')
    def get_metadata(self, share_network, share_network_subnet):
        return super().get_metadata(
            share_network, subresource=share_network_subnet
        )

    @api_versions.wraps('2.78')
    def set_metadata(self, resource, metadata, subresource=None):
        return super().set_metadata(
            resource, metadata, subresource=subresource
        )

    @api_versions.wraps('2.78')
    def delete_metadata(self, resource, keys, subresource=None):
        return super().delete_metadata(resource, keys, subresource=subresource)

    @api_versions.wraps('2.78')
    def update_all_metadata(self, resource, metadata, subresource=None):
        return super().update_all_metadata(
            resource, metadata, subresource=subresource
        )



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base

RESOURCE_PATH_LEGACY = '/os-quota-sets'
RESOURCE_PATH = '/quota-sets'
REPLICA_QUOTAS_MICROVERSION = "2.53"


class QuotaSet(base.Resource):
    @property
    def id(self):
        """Needed by Resource to self-refresh and be indexed."""
        return self.tenant_id

    def update(self, *args, **kwargs):
        self.manager.update(self.tenant_id, *args, **kwargs)


class QuotaSetManager(base.ManagerWithFind):
    resource_class = QuotaSet

    def _check_user_id_and_share_type_args(self, user_id, share_type):
        if user_id and share_type:
            raise ValueError(
                "'user_id' and 'share_type' values are mutually exclusive. "
                "one or both should be unset."
            )

    def _do_get(
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        detail=False,
        resource_path=RESOURCE_PATH,
    ):
        self._check_user_id_and_share_type_args(user_id, share_type)
        if hasattr(tenant_id, 'tenant_id'):
            tenant_id = tenant_id.tenant_id

        if detail:
            query = '/detail'
        else:
            query = ''

        if user_id and share_type:
            query = f'{query}?user_id={user_id}&share_type={share_type}'
        elif user_id:
            query = f'{query}?user_id={user_id}'
        elif share_type:
            query = f'{query}?share_type={share_type}'
        data = {
            "resource_path": resource_path,
            "tenant_id": tenant_id,
        }

        url = ("%(resource_path)s/%(tenant_id)s" + query) % data
        return self._get(url, "quota_set")

    @api_versions.wraps("1.0", "2.6")
    def get(self, tenant_id, user_id=None, detail=False):
        return self._do_get(
            tenant_id, user_id, resource_path=RESOURCE_PATH_LEGACY
        )

    @api_versions.wraps("2.7", "2.24")  # noqa
    def get(self, tenant_id, user_id=None, detail=False):  # noqa
        return self._do_get(tenant_id, user_id, resource_path=RESOURCE_PATH)

    @api_versions.wraps("2.25", "2.38")  # noqa
    def get(self, tenant_id, user_id=None, detail=False):  # noqa
        return self._do_get(
            tenant_id, user_id, detail=detail, resource_path=RESOURCE_PATH
        )

    @api_versions.wraps("2.39")  # noqa
    def get(self, tenant_id, user_id=None, share_type=None, detail=False):  # noqa
        return self._do_get(
            tenant_id,
            user_id,
            share_type=share_type,
            detail=detail,
            resource_path=RESOURCE_PATH,
        )

    def _do_update(
        self,
        tenant_id,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        force=None,
        user_id=None,
        share_type=None,
        share_groups=None,
        share_group_snapshots=None,
        share_replicas=None,
        replica_gigabytes=None,
        per_share_gigabytes=None,
        encryption_keys=None,
        resource_path=RESOURCE_PATH,
    ):
        self._check_user_id_and_share_type_args(user_id, share_type)
        body = {
            'quota_set': {
                'tenant_id': tenant_id,
                'shares': shares,
                'snapshots': snapshots,
                'gigabytes': gigabytes,
                'snapshot_gigabytes': snapshot_gigabytes,
                'share_networks': share_networks,
                'share_groups': share_groups,
                'share_group_snapshots': share_group_snapshots,
                'force': force,
                'share_replicas': share_replicas,
                'replica_gigabytes': replica_gigabytes,
                'per_share_gigabytes': per_share_gigabytes,
                'encryption_keys': encryption_keys,
            },
        }

        for key in list(body['quota_set']):
            if body['quota_set'][key] is None:
                body['quota_set'].pop(key)
        data = {
            "resource_path": resource_path,
            "tenant_id": tenant_id,
            "user_id": user_id,
            "st": share_type,
        }
        if user_id:
            url = '{resource_path}/{tenant_id}?user_id={user_id}'.format(
                **data
            )
        elif share_type:
            url = '{resource_path}/{tenant_id}?share_type={st}'.format(**data)
        else:
            url = "{resource_path}/{tenant_id}".format(**data)

        return self._update(url, body, 'quota_set')

    @api_versions.wraps("1.0", "2.6")
    def update(
        self,
        tenant_id,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        force=None,
        user_id=None,
    ):
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            resource_path=RESOURCE_PATH_LEGACY,
        )

    @api_versions.wraps("2.7", "2.38")  # noqa
    def update(  # noqa
        self,
        tenant_id,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        force=None,
        user_id=None,
    ):
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            resource_path=RESOURCE_PATH,
        )

    def _validate_st_and_sn_in_same_request(self, share_type, share_networks):
        if share_type and share_networks:
            raise ValueError(
                "'share_networks' quota can be set only for project or user, "
                "not share type."
            )

    @api_versions.wraps("2.39", "2.39")  # noqa
    def update(  # noqa
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        force=None,
    ):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            share_type=share_type,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps("2.40", "2.52")  # noqa
    def update(  # noqa
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        share_groups=None,
        share_group_snapshots=None,
        force=None,
    ):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps(REPLICA_QUOTAS_MICROVERSION, "2.61")  # noqa
    def update(  # noqa
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        share_groups=None,
        share_group_snapshots=None,
        share_replicas=None,
        replica_gigabytes=None,
        force=None,
    ):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            share_replicas=share_replicas,
            replica_gigabytes=replica_gigabytes,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps("2.62", "2.89")  # noqa
    def update(  # noqa
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        share_groups=None,
        share_group_snapshots=None,
        share_replicas=None,
        replica_gigabytes=None,
        force=None,
        per_share_gigabytes=None,
    ):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            share_replicas=share_replicas,
            replica_gigabytes=replica_gigabytes,
            per_share_gigabytes=per_share_gigabytes,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps("2.90")  # noqa
    def update(  # noqa
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        shares=None,
        snapshots=None,
        gigabytes=None,
        snapshot_gigabytes=None,
        share_networks=None,
        share_groups=None,
        share_group_snapshots=None,
        share_replicas=None,
        replica_gigabytes=None,
        force=None,
        per_share_gigabytes=None,
        encryption_keys=None,
    ):
        self._validate_st_and_sn_in_same_request(share_type, share_networks)
        return self._do_update(
            tenant_id,
            shares,
            snapshots,
            gigabytes,
            snapshot_gigabytes,
            share_networks,
            force,
            user_id,
            share_type=share_type,
            share_groups=share_groups,
            share_group_snapshots=share_group_snapshots,
            share_replicas=share_replicas,
            replica_gigabytes=replica_gigabytes,
            per_share_gigabytes=per_share_gigabytes,
            encryption_keys=encryption_keys,
            resource_path=RESOURCE_PATH,
        )

    @api_versions.wraps("1.0", "2.6")
    def defaults(self, tenant_id):
        return self._get(
            f"{RESOURCE_PATH_LEGACY}/{tenant_id}/defaults", "quota_set"
        )

    @api_versions.wraps("2.7")  # noqa
    def defaults(self, tenant_id):  # noqa
        return self._get(f"{RESOURCE_PATH}/{tenant_id}/defaults", "quota_set")

    def _do_delete(
        self,
        tenant_id,
        user_id=None,
        share_type=None,
        resource_path=RESOURCE_PATH,
    ):
        self._check_user_id_and_share_type_args(user_id, share_type)
        data = {
            "resource_path": resource_path,
            "tenant_id": tenant_id,
            "user_id": user_id,
            "st": share_type,
        }
        if user_id:
            url = '{resource_path}/{tenant_id}?user_id={user_id}'.format(
                **data
            )
        elif share_type:
            url = '{resource_path}/{tenant_id}?share_type={st}'.format(**data)
        else:
            url = '{resource_path}/{tenant_id}'.format(**data)
        self._delete(url)

    @api_versions.wraps("1.0", "2.6")
    def delete(self, tenant_id, user_id=None):
        return self._do_delete(
            tenant_id, user_id, resource_path=RESOURCE_PATH_LEGACY
        )

    @api_versions.wraps("2.7", "2.38")  # noqa
    def delete(self, tenant_id, user_id=None):  # noqa
        return self._do_delete(tenant_id, user_id, resource_path=RESOURCE_PATH)

    @api_versions.wraps("2.39")  # noqa
    def delete(self, tenant_id, user_id=None, share_type=None):  # noqa
        return self._do_delete(
            tenant_id, user_id, share_type, resource_path=RESOURCE_PATH
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from manilaclient.tests.functional.osc import base


class ShareSnapshotInstancesCLITest(base.OSCClientTestBase):
    def test_openstack_share_snapshot_instance_list(self):
        share = self.create_share()
        snapshot = self.create_snapshot(share['id'])
        share_snapshot_instances_list = self.listing_result(
            'share snapshot instance', 'list --detailed'
        )
        self.assertTableStruct(
            share_snapshot_instances_list,
            [
                'ID',
                'Snapshot ID',
                'Status',
                'Created At',
                'Updated At',
                'Share ID',
                'Share Instance ID',
                'Progress',
                'Provider Location',
            ],
        )
        self.assertIn(
            snapshot['id'],
            [item['Snapshot ID'] for item in (share_snapshot_instances_list)],
        )

    def test_openstack_share_snapshot_instance_show(self):
        share = self.create_share()
        snapshot = self.create_snapshot(share['id'], wait=True)

        share_snapshot_instance = self.listing_result(
            "share snapshot instance", f'list --snapshot {snapshot["id"]}'
        )
        result = self.dict_result(
            'share snapshot instance',
            f'show {share_snapshot_instance[0]["ID"]}',
        )

        self.assertEqual(share_snapshot_instance[0]['ID'], result['id'])
        self.assertEqual(
            share_snapshot_instance[0]['Snapshot ID'], result['snapshot_id']
        )

        listing_result = self.listing_result(
            'share snapshot instance',
            f'show {share_snapshot_instance[0]["ID"]}',
        )
        self.assertTableStruct(listing_result, ['Field', 'Value'])

    def test_openstack_share_snapshot_instance_set(self):
        share = self.create_share()
        snapshot = self.create_snapshot(share['id'], wait=True)
        share_snapshot_instance = self.listing_result(
            "share snapshot instance", f'list --snapshot {snapshot["id"]}'
        )
        result1 = self.dict_result(
            'share snapshot instance',
            f'show {share_snapshot_instance[0]["ID"]}',
        )
        self.assertEqual(share_snapshot_instance[0]['ID'], result1['id'])
        self.assertEqual(snapshot['id'], result1['snapshot_id'])
        self.assertEqual('available', result1['status'])

        self.openstack(
            'share snapshot instance set '
            f'{share_snapshot_instance[0]["ID"]} --status error'
        )
        result2 = self.dict_result(
            'share snapshot instance',
            f'show {share_snapshot_instance[0]["ID"]}',
        )
        self.assertEqual(share_snapshot_instance[0]["ID"], result2['id'])
        self.assertEqual('error', result2['status'])
